from __future__ import annotations

from typing_extensions import override


"""Generic persistent, concurrent dictionary-like facility."""


__copyright__ = """
Copyright (C) 2011,2014 Andreas Kloeckner
Copyright (C) 2017 Matt Wala
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import logging
import os
import pickle
import sqlite3
import sys
from collections.abc import Callable, Iterator, Mapping
from dataclasses import fields as dc_fields, is_dataclass
from enum import Enum
from typing import Any, TypeVar, cast
from warnings import warn

from siphash24 import siphash13

from pytools import Hash  # noqa: TC001 # Some places are importing it from here.


class RecommendedHashNotFoundWarning(UserWarning):
    pass


try:
    import attrs
except ModuleNotFoundError:
    _HAS_ATTRS = False
else:
    _HAS_ATTRS = True


# Used to speed up the comparison in update_for_numpy_scalar and for mock testing
_IS_BIGENDIAN = sys.byteorder == "big"


logger = logging.getLogger(__name__)

# NOTE: not always available so they get hardcoded here
SQLITE_BUSY = getattr(sqlite3, "SQLITE_BUSY", 5)
SQLITE_CONSTRAINT_PRIMARYKEY = getattr(sqlite3, "SQLITE_CONSTRAINT_PRIMARYKEY", 1555)

__doc__ = """
Persistent Hashing and Persistent Dictionaries
==============================================

This module contains functionality that allows hashing with keys that remain
valid across interpreter invocations, unlike Python's built-in hashes.

This module also provides a disk-backed dictionary that uses persistent hashing.

.. autoexception:: NoSuchEntryError
.. autoexception:: NoSuchEntryCollisionError
.. autoexception:: ReadOnlyEntryError

.. autoexception:: CollisionWarning
.. autoexception:: NanKeyWarning

.. autoclass:: KeyBuilder
.. autoclass:: PersistentDict
.. autoclass:: WriteOncePersistentDict


Internal stuff that is only here because the documentation tool wants it
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. class:: K

    A type variable for the key type of a :class:`PersistentDict`.

.. class:: V

    A type variable for the value type of a :class:`PersistentDict`.
"""


# {{{ key generation

class NanKeyWarning(UserWarning):
    """Warning raised when a NaN is encountered while hashing a key in
    :class:`KeyBuilder`."""


class KeyBuilder:
    """A (stateless) object that computes persistent hashes of objects fed to it.
    Subclassing this class permits customizing the computation of hash keys.

    This class follows the same general rules as Python's built-in hashing:

    - Only immutable objects can be hashed.
    - If two objects compare equal, they must hash to the same value.
    - Objects with the same hash may or may not compare equal.

    In addition, hashes computed with :class:`KeyBuilder` have the following
    properties:

    - The hash is persistent across interpreter invocations.
    - The hash is the same across different Python versions and platforms.
    - The hash is invariant with respect to :envvar:`PYTHONHASHSEED`.
    - Hashes are computed using functionality from :mod:`hashlib`.

    Key builders of this type are used by :class:`PersistentDict`, but
    other uses are entirely allowable.

    .. automethod:: __call__
    .. automethod:: rec
    .. staticmethod:: new_hash()

        Return a new hash instance following the protocol of the ones
        from :mod:`hashlib`. This will permit switching to different
        hash algorithms in the future. Subclasses are expected to use
        this to create new hashes. Not doing so is deprecated and
        may stop working as early as 2022.

        .. versionadded:: 2021.2
    """

    # this exists so that we can (conceivably) switch algorithms at some point
    # down the road
    new_hash: Callable[..., Hash] = siphash13

    def rec(self, key_hash: Hash, key: Any) -> Hash:
        """
        :arg key_hash: the hash object to be updated with the hash of *key*.
        :arg key: the (immutable) Python object to be hashed.
        :returns: the updated *key_hash*

        .. versionchanged:: 2021.2

            Now returns the updated *key_hash*.
        """

        digest = getattr(key, "_pytools_persistent_hash_digest", None)

        if digest is None and not isinstance(key, type):
            try:
                method = key.update_persistent_hash
            except AttributeError:
                pass
            else:
                inner_key_hash = self.new_hash()
                method(inner_key_hash, self)
                digest = inner_key_hash.digest()

        if digest is None:
            tp = type(key)
            tname = tp.__name__
            method = None
            try:
                method = getattr(self, "update_for_"+tname)
            except AttributeError:
                if "numpy" in sys.modules:
                    import numpy as np

                    # Hashing numpy dtypes
                    if (
                            # Handling numpy >= 1.20, for which
                            # type(np.dtype("float32")) -> "dtype[float32]"
                            tname.startswith("dtype[")
                            # Handling numpy >= 1.25, for which
                            # type(np.dtype("float32")) -> "Float32DType"
                            or tname.endswith("DType")
                            ):
                        if isinstance(key, np.dtype):
                            method = self.update_for_specific_dtype

                    # Hashing numpy scalars
                    elif isinstance(key, np.number | np.bool_):
                        # Non-numpy scalars are handled above in the try block.
                        method = self.update_for_numpy_scalar

                    # numpy arrays are mutable so they are not handled in KeyBuilder

                if method is None:
                    if issubclass(tp, Enum):
                        method = self.update_for_enum

                    elif is_dataclass(tp):
                        method = self.update_for_dataclass

                    elif _HAS_ATTRS and attrs.has(tp):
                        method = self.update_for_attrs

            if method is not None:
                inner_key_hash = self.new_hash()
                method(inner_key_hash, key)
                digest = inner_key_hash.digest()

        if digest is None:
            raise TypeError(
                    f"unsupported type for persistent hash keying: {type(key)}")

        if not isinstance(key, type):
            try:
                object.__setattr__(key, "_pytools_persistent_hash_digest", digest)
            except AttributeError:
                pass
            except TypeError:
                pass

        key_hash.update(digest)
        return key_hash

    def __call__(self, key: Any) -> str:
        """Return the hash of *key*."""
        key_hash = self.new_hash()
        self.rec(key_hash, key)
        return key_hash.hexdigest()

    # {{{ updaters

    # NOTE: None of these should be static or classmethods. While Python itself is
    # perfectly OK with overriding those with 'normal' methods, type checkers
    # understandably don't like it.

    def update_for_type(self, key_hash: Hash, key: type) -> None:
        key_hash.update(
            f"{key.__module__}.{key.__qualname__}.{key.__name__}".encode())

    update_for_ABCMeta = update_for_type

    def update_for_int(self, key_hash: Hash, key: int) -> None:
        sz = 8
        # Note: this must match the hash for numpy integers, since
        # np.int64(1) == 1
        while True:
            try:
                key_hash.update(key.to_bytes(sz, byteorder="little", signed=True))
                return
            except OverflowError:
                sz *= 2

    def update_for_enum(self, key_hash: Hash, key: Enum) -> None:
        self.update_for_str(key_hash, str(key))

    def update_for_bool(self, key_hash: Hash, key: bool) -> None:
        key_hash.update(str(key).encode("utf8"))

    def update_for_float(self, key_hash: Hash, key: float) -> None:
        import math
        if math.isnan(key):
            # Also applies to np.nan, float("nan")
            warn("Encountered a NaN while hashing. Since NaNs compare unequal "
                 "to themselves, the resulting key can not be retrieved from a "
                 "PersistentDict and will lead to a collision error on retrieval.",
                 NanKeyWarning,
                 stacklevel=1)

        key_hash.update(key.hex().encode("utf8"))

    def update_for_complex(self, key_hash: Hash, key: float) -> None:
        key_hash.update(repr(key).encode("utf-8"))

    def update_for_str(self, key_hash: Hash, key: str) -> None:
        key_hash.update(key.encode("utf8"))

    def update_for_bytes(self, key_hash: Hash, key: bytes) -> None:
        key_hash.update(key)

    def update_for_tuple(self, key_hash: Hash, key: tuple[Any, ...]) -> None:
        for obj_i in key:
            self.rec(key_hash, obj_i)

    def update_for_frozenset(self, key_hash: Hash, key: frozenset[Any]) -> None:
        from pytools import unordered_hash

        unordered_hash(
            key_hash,
            (self.rec(self.new_hash(), key_i).digest() for key_i in key),
            hash_constructor=self.new_hash)

    update_for_FrozenOrderedSet = update_for_frozenset

    def update_for_NoneType(self, key_hash: Hash, key: None) -> None:
        del key
        key_hash.update(b"<None>")

    def update_for_dtype(self, key_hash: Hash, key: Any) -> None:
        key_hash.update(key.str.encode("utf8"))

    # Handling numpy >= 1.20, for which
    # type(np.dtype("float32")) -> "dtype[float32]"
    # Introducing this method allows subclasses to specially handle all those
    # dtypes.
    def update_for_specific_dtype(self, key_hash: Hash, key: Any) -> None:
        key_hash.update(key.str.encode("utf8"))

    def update_for_numpy_scalar(self, key_hash: Hash, key: Any) -> None:
        import numpy as np

        # tobytes() of np.complex256 and np.float128 are non-deterministic,
        # so we need to use their repr() instead.
        if hasattr(np, "complex256") and key.dtype == np.dtype("complex256"):
            key_hash.update(repr(complex(key)).encode("utf8"))
        elif hasattr(np, "float128") and key.dtype == np.dtype("float128"):
            key_hash.update(repr(float(key)).encode("utf8"))
        else:
            if _IS_BIGENDIAN:
                key_hash.update(np.array(key).byteswap().tobytes())
            else:
                key_hash.update(np.array(key).tobytes())

    def update_for_dataclass(self, key_hash: Hash, key: Any) -> None:
        self.rec(key_hash, f"{type(key).__qualname__}.{type(key).__name__}")

        for fld in dc_fields(key):
            self.rec(key_hash, fld.name)
            self.rec(key_hash, getattr(key, fld.name, None))

    def update_for_attrs(self, key_hash: Hash, key: Any) -> None:
        self.rec(key_hash, f"{type(key).__qualname__}.{type(key).__name__}")

        for fld in attrs.fields(key.__class__):
            self.rec(key_hash, fld.name)
            self.rec(key_hash, getattr(key, fld.name, None))

    def update_for_frozendict(self, key_hash: Hash, key: Mapping[Any, Any]) -> None:
        from pytools import unordered_hash

        unordered_hash(
            key_hash,
            (self.rec(self.new_hash(), (k, v)).digest() for k, v in key.items()),
            hash_constructor=self.new_hash)

    update_for_immutabledict = update_for_frozendict
    update_for_constantdict = update_for_frozendict
    update_for_PMap = update_for_frozendict
    update_for_Map = update_for_frozendict

    # {{{ date, time, datetime, timezone

    def update_for_date(self, key_hash: Hash, key: Any) -> None:
        # 'date' has no timezone information; it is always naive
        self.rec(key_hash, key.isoformat())

    def update_for_time(self, key_hash: Hash, key: Any) -> None:
        # 'time' should differentiate between naive and aware
        import datetime

        # Convert to datetime object
        self.rec(key_hash, datetime.datetime.combine(datetime.date.min, key))
        self.rec(key_hash, "<time>")

    def update_for_datetime(self, key_hash: Hash, key: Any) -> None:
        # 'datetime' should differentiate between naive and aware

        # https://docs.python.org/3.11/library/datetime.html#determining-if-an-object-is-aware-or-naive
        if key.tzinfo is not None and key.tzinfo.utcoffset(key) is not None:
            self.rec(key_hash, key.timestamp())
            self.rec(key_hash, "<aware>")
        else:
            from datetime import timezone
            self.rec(key_hash, key.replace(tzinfo=timezone.utc).timestamp())
            self.rec(key_hash, "<naive>")

    def update_for_timezone(self, key_hash: Hash, key: Any) -> None:
        self.rec(key_hash, repr(key))

    # }}}

    def update_for_function(self, key_hash: Hash, key: Any) -> None:
        self.rec(key_hash, key.__module__ + key.__qualname__)

        if key.__closure__:
            self.rec(key_hash, tuple(c.cell_contents for c in key.__closure__))

    # }}}

# }}}


# {{{ top-level

class NoSuchEntryError(KeyError):
    """Raised when an entry is not found in a :class:`PersistentDict`."""


class NoSuchEntryCollisionError(NoSuchEntryError):
    """Raised when an entry is not found in a :class:`PersistentDict`, but it
    contains an entry with the same hash key (hash collision)."""


class ReadOnlyEntryError(KeyError):
    """Raised when an attempt is made to overwrite an entry in a
    :class:`WriteOncePersistentDict`."""


class CollisionWarning(UserWarning):
    """Warning raised when a collision is detected in a :class:`PersistentDict`."""


def __getattr__(name: str) -> Any:
    if name in ("NoSuchEntryInvalidKeyError",
                "NoSuchEntryInvalidContentsError"):
        warn(f"pytools.persistent_dict.{name} has been removed.", stacklevel=2)
        return NoSuchEntryError

    raise AttributeError(name)


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class _PersistentDictBase(Mapping[K, V]):
    def __init__(self,
                 identifier: str,
                 key_builder: KeyBuilder | None = None,
                 container_dir: str | None = None,
                 enable_wal: bool = False,
                 safe_sync: bool | None = None) -> None:
        self.identifier = identifier
        self.conn = None

        if key_builder is None:
            key_builder = KeyBuilder()

        self.key_builder = key_builder

        from os.path import join
        if container_dir is None:
            import platformdirs

            if sys.platform == "darwin" and os.getenv("XDG_CACHE_HOME") is not None:
                # platformdirs does not handle XDG_CACHE_HOME on macOS
                # https://github.com/platformdirs/platformdirs/issues/269
                container_dir = join(os.getenv("XDG_CACHE_HOME"), "pytools")
            else:
                container_dir = platformdirs.user_cache_dir("pytools", "pytools")

        self.filename = join(container_dir, f"pdict-v5-{identifier}-"
                             + ".".join(str(i) for i in sys.version_info)
                             + ".sqlite")

        self.container_dir = container_dir
        self._make_container_dir()

        from threading import Lock
        self.mutex = Lock()

        # * isolation_level=None: enable autocommit mode
        #   https://www.sqlite.org/lang_transaction.html#implicit_versus_explicit_transactions
        # * check_same_thread=False: thread-level concurrency is handled by the
        #   mutex above
        self.conn = sqlite3.connect(self.filename,
                                    isolation_level=None,
                                    check_same_thread=False)

        self._exec_sql(
            "CREATE TABLE IF NOT EXISTS dict "
            "(keyhash TEXT NOT NULL PRIMARY KEY, key_value TEXT NOT NULL)"
            )

        # https://www.sqlite.org/wal.html
        if enable_wal:
            self._exec_sql("PRAGMA journal_mode = 'WAL'")

        # Note: the following configuration values were taken mostly from litedict:
        # https://github.com/litements/litedict/blob/377603fa597453ffd9997186a493ed4fd23e5399/litedict.py#L67-L70

        # Use in-memory temp store
        # https://www.sqlite.org/pragma.html#pragma_temp_store
        self._exec_sql("PRAGMA temp_store = 'MEMORY'")

        # fsync() can be extremely slow on some systems.
        # See https://github.com/inducer/pytools/issues/227 for context.
        # https://www.sqlite.org/pragma.html#pragma_synchronous
        if safe_sync is None or safe_sync:
            if safe_sync is None:
                warn(f"pytools.persistent_dict '{identifier}': "
                     "enabling safe_sync as default. "
                     "This provides strong protection against data loss, "
                     "but can be unnecessarily expensive for use cases such as "
                     "caches."
                     "Pass 'safe_sync=False' if occasional data loss is tolerable. "
                     "Pass 'safe_sync=True' to suppress this warning.",
                     stacklevel=3)
            self._exec_sql("PRAGMA synchronous = 'NORMAL'")
        else:
            self._exec_sql("PRAGMA synchronous = 'OFF'")

        # 64 MByte of cache
        # https://www.sqlite.org/pragma.html#pragma_cache_size
        self._exec_sql("PRAGMA cache_size = -64000")

    def __del__(self) -> None:
        with self.mutex:
            if self.conn:
                self.conn.close()

    def _collision_check(self, key: K, stored_key: K) -> None:
        if stored_key != key:
            # Key collision, oh well.
            warn(f"{self.identifier}: key collision in cache at "
                    f"'{self.container_dir}' -- these are sufficiently unlikely "
                    "that they're often indicative of a broken hash key "
                    "implementation (that is not considering some elements "
                    "relevant for equality comparison)",
                    CollisionWarning,
                    stacklevel=3
                 )

            # This is here so we can step through equality comparison to
            # see what is actually non-equal.
            stored_key == key  # noqa: B015
            raise NoSuchEntryCollisionError(key)

    def _exec_sql(self, *args: Any) -> sqlite3.Cursor:
        def execute() -> sqlite3.Cursor:
            assert self.conn is not None
            return self.conn.execute(*args)

        cursor = self._exec_sql_fn(execute)
        if not isinstance(cursor, sqlite3.Cursor):
            raise RuntimeError("Failed to execute SQL statement")

        return cursor

    def _exec_sql_fn(self, fn: Callable[[], T]) -> T | None:
        n = 0

        with self.mutex:
            while True:
                n += 1
                try:
                    return fn()
                except sqlite3.OperationalError as e:
                    # If the database is busy, retry
                    if (hasattr(e, "sqlite_errorcode")
                        and e.sqlite_errorcode != SQLITE_BUSY):
                        raise
                    if n % 20 == 0:
                        warn(f"PersistentDict: database '{self.filename}' busy, {n} "
                             "retries", stacklevel=3)
                else:
                    break

    def store_if_not_present(self, key: K, value: V) -> None:
        """Store (*key*, *value*) if *key* is not already present."""
        self.store(key, value, _skip_if_present=True)

    def store(self, key: K, value: V, _skip_if_present: bool = False) -> None:
        """Store (*key*, *value*) in the dictionary."""
        raise NotImplementedError

    def fetch(self, key: K) -> V:
        """Return the value associated with *key* in the dictionary."""
        raise NotImplementedError

    def _make_container_dir(self) -> None:
        """Create the container directory to store the dictionary."""
        os.makedirs(self.container_dir, exist_ok=True)

    @override
    def __getitem__(self, key: K) -> V:
        """Return the value associated with *key* in the dictionary."""
        return self.fetch(key)

    def __setitem__(self, key: K, value: V) -> None:
        """Store (*key*, *value*) in the dictionary."""
        self.store(key, value)

    @override
    def __len__(self) -> int:
        """Return the number of entries in the dictionary."""
        result, = next(self._exec_sql("SELECT COUNT(*) FROM dict"))
        assert isinstance(result, int)
        return result

    @override
    def __iter__(self) -> Iterator[K]:
        """Return an iterator over the keys in the dictionary."""
        return self.keys()

    @override
    def keys(self) -> Iterator[K]:
        """Return an iterator over the keys in the dictionary."""
        for row in self._exec_sql("SELECT key_value FROM dict ORDER BY rowid"):
            yield pickle.loads(row[0])[0]

    @override
    def values(self) -> Iterator[V]:
        """Return an iterator over the values in the dictionary."""
        for row in self._exec_sql("SELECT key_value FROM dict ORDER BY rowid"):
            yield pickle.loads(row[0])[1]

    @override
    def items(self) -> Iterator[tuple[K, V]]:
        """Return an iterator over the items in the dictionary."""
        for row in self._exec_sql("SELECT key_value FROM dict ORDER BY rowid"):
            yield pickle.loads(row[0])

    def nbytes(self) -> int:
        """Return the size of the dictionary in bytes."""
        result, = next(self._exec_sql("SELECT page_size * page_count FROM "
                                      "pragma_page_size(), pragma_page_count()"))
        assert isinstance(result, int)

        return result

    @override
    def __repr__(self) -> str:
        """Return a string representation of the dictionary."""
        return f"{type(self).__name__}({self.filename}, nitems={len(self)})"

    def clear(self) -> None:
        """Remove all entries from the dictionary."""
        self._exec_sql("DELETE FROM dict")


class WriteOncePersistentDict(_PersistentDictBase[K, V]):
    """A concurrent disk-backed dictionary that disallows overwriting/
    deletion (but allows removing all entries).

    Compared with :class:`PersistentDict`, this class has faster
    retrieval times because it uses an LRU cache to cache entries in memory.

    .. note::

        This class intentionally does not store all values with a certain
        key, based on the assumption that key conflicts are highly unlikely,
        and if they occur, almost always due to a bug in the hash key
        generation code (:class:`KeyBuilder`).

    .. warning::

        Since NaNs compare unequal to themselves, keys that include NaNs can
        not be retrieved from a :class:`WriteOncePersistentDict` and will lead to a
        :exc:`NoSuchEntryCollisionError` on retrieval.

    .. automethod:: __init__
    .. automethod:: __getitem__
    .. automethod:: __setitem__
    .. automethod:: clear
    .. automethod:: clear_in_mem_cache
    .. automethod:: store
    .. automethod:: store_if_not_present
    .. automethod:: fetch
    """
    def __init__(self, identifier: str,
                 key_builder: KeyBuilder | None = None,
                 container_dir: str | None = None,
                 *,
                 enable_wal: bool = False,
                 safe_sync: bool | None = None,
                 in_mem_cache_size: int = 256) -> None:
        """
        :arg identifier: a filename-compatible string identifying this
            dictionary
        :arg key_builder: a subclass of :class:`KeyBuilder`
        :arg container_dir: the directory in which to store this
            dictionary. If ``None``, the default cache directory from
            :func:`platformdirs.user_cache_dir` is used
        :arg enable_wal: enable write-ahead logging (WAL) mode. This mode
            is faster than the default rollback journal mode, but it is
            not compatible with network filesystems.
        :arg in_mem_cache_size: retain an in-memory cache of up to
            *in_mem_cache_size* items (with an LRU replacement policy)
        """
        super().__init__(identifier,
                         key_builder=key_builder,
                         container_dir=container_dir,
                         enable_wal=enable_wal,
                         safe_sync=safe_sync)

        from functools import lru_cache

        self._fetch = lru_cache(maxsize=in_mem_cache_size)(self._fetch_uncached)

    def clear_in_mem_cache(self) -> None:
        """
        Clear the in-memory cache of this dictionary.

        .. versionadded:: 2023.1.1
        """
        self._fetch.cache_clear()

    @override
    def store(self, key: K, value: V, _skip_if_present: bool = False) -> None:
        keyhash = self.key_builder(key)
        v = pickle.dumps((key, value))

        if _skip_if_present:
            self._exec_sql("INSERT OR IGNORE INTO dict VALUES (?, ?)",
                              (keyhash, v))
        else:
            try:
                self._exec_sql("INSERT INTO dict VALUES (?, ?)", (keyhash, v))
            except sqlite3.IntegrityError as e:
                if hasattr(e, "sqlite_errorcode"):
                    if e.sqlite_errorcode == SQLITE_CONSTRAINT_PRIMARYKEY:
                        raise ReadOnlyEntryError("WriteOncePersistentDict, "
                                                 "tried overwriting key") from e
                    raise
                raise ReadOnlyEntryError("WriteOncePersistentDict, "
                                         "tried overwriting key") from e

    def _fetch_uncached(self, keyhash: str) -> tuple[K, V]:
        # This method is separate from fetch() to allow for LRU caching

        def fetch_inner() -> tuple[Any] | None:
            assert self.conn is not None

            # This is separate from fetch() so that the mutex covers the
            # fetchone() call
            c = self.conn.execute("SELECT key_value FROM dict WHERE keyhash=?",
                                (keyhash,))
            res = c.fetchone()
            assert res is None or isinstance(res, tuple)
            return res

        row = self._exec_sql_fn(fetch_inner)
        if row is None:
            raise KeyError

        key, value = pickle.loads(row[0])
        return key, value

    @override
    def fetch(self, key: K) -> V:
        keyhash = self.key_builder(key)

        try:
            stored_key, value = self._fetch(keyhash)
        except KeyError as err:
            raise NoSuchEntryError(key) from err
        else:
            self._collision_check(key, stored_key)
            return value

    @override
    def clear(self) -> None:
        super().clear()
        self._fetch.cache_clear()


class PersistentDict(_PersistentDictBase[K, V]):
    """A concurrent disk-backed dictionary.

    .. note::

        This class intentionally does not store all values with a certain
        key, based on the assumption that key conflicts are highly unlikely,
        and if they occur, almost always due to a bug in the hash key
        generation code (:class:`KeyBuilder`).

    .. warning::

        Since NaNs compare unequal to themselves, keys that include NaNs can
        not be retrieved from a :class:`PersistentDict` and will lead to a
        :exc:`NoSuchEntryCollisionError` on retrieval.

    .. automethod:: __init__
    .. automethod:: __getitem__
    .. automethod:: __setitem__
    .. automethod:: __delitem__
    .. automethod:: clear
    .. automethod:: store
    .. automethod:: store_if_not_present
    .. automethod:: fetch
    .. automethod:: remove
    """
    def __init__(self,
                 identifier: str,
                 key_builder: KeyBuilder | None = None,
                 container_dir: str | None = None,
                 *,
                 enable_wal: bool = False,
                 safe_sync: bool | None = None) -> None:
        """
        :arg identifier: a filename-compatible string identifying this
            dictionary
        :arg key_builder: a subclass of :class:`KeyBuilder`
        :arg container_dir: the directory in which to store this
            dictionary. If ``None``, the default cache directory from
            :func:`platformdirs.user_cache_dir` is used
        :arg enable_wal: enable write-ahead logging (WAL) mode. This mode
            is faster than the default rollback journal mode, but it is
            not compatible with network filesystems.
        """
        super().__init__(identifier,
                         key_builder=key_builder,
                         container_dir=container_dir,
                         enable_wal=enable_wal,
                         safe_sync=safe_sync)

    @override
    def store(self, key: K, value: V, _skip_if_present: bool = False) -> None:
        keyhash = self.key_builder(key)
        v = pickle.dumps((key, value))

        mode = "IGNORE" if _skip_if_present else "REPLACE"

        self._exec_sql(f"INSERT OR {mode} INTO dict VALUES (?, ?)",
                              (keyhash, v))

    @override
    def fetch(self, key: K) -> V:
        keyhash = self.key_builder(key)

        def fetch_inner() -> tuple[Any] | None:
            assert self.conn is not None

            # This is separate from fetch() so that the mutex covers the
            # fetchone() call
            c = self.conn.execute("SELECT key_value FROM dict WHERE keyhash=?",
                                (keyhash,))
            res = c.fetchone()
            assert res is None or isinstance(res, tuple)
            return res

        row = self._exec_sql_fn(fetch_inner)

        if row is None:
            raise NoSuchEntryError(key)

        stored_key, value = pickle.loads(row[0])
        self._collision_check(key, stored_key)
        return cast("V", value)

    def remove(self, key: K) -> None:
        """Remove the entry associated with *key* from the dictionary."""
        keyhash = self.key_builder(key)

        def remove_inner() -> None:
            assert self.conn is not None
            self.conn.execute("BEGIN EXCLUSIVE TRANSACTION")
            try:
                # This is split into SELECT/DELETE to allow for a collision check
                c = self.conn.execute("SELECT key_value FROM dict WHERE "
                                        "keyhash=?", (keyhash,))
                row = c.fetchone()
                if row is None:
                    raise NoSuchEntryError(key)

                stored_key, _value = pickle.loads(row[0])
                self._collision_check(key, stored_key)

                self.conn.execute("DELETE FROM dict WHERE keyhash=?", (keyhash,))
                self.conn.execute("COMMIT")
            except Exception as e:
                self.conn.execute("ROLLBACK")
                raise e

        self._exec_sql_fn(remove_inner)

    def __delitem__(self, key: K) -> None:
        """Remove the entry associated with *key* from the dictionary."""
        self.remove(key)

# }}}

# vim: foldmethod=marker



"""
.. autofunction:: estimate_order_of_convergence
.. autoclass:: EOCRecorder
.. autofunction:: stringify_eocs
.. autoclass:: PConvergenceVerifier
"""

from __future__ import annotations

import numbers

import numpy as np
from typing_extensions import override


# {{{ eoc estimation --------------------------------------------------------------

def estimate_order_of_convergence(abscissae, errors):
    r"""Assuming that abscissae and errors are connected by a law of the form

    .. math::

        \text{Error} =  \text{constant} \cdot \text{abscissa }^{\text{order}},

    this function finds, in a least-squares sense, the best approximation of
    constant and order for the given data set. It returns a tuple (constant, order).
    """
    assert len(abscissae) == len(errors)
    if len(abscissae) <= 1:
        raise RuntimeError("Need more than one value to guess order of convergence.")

    coefficients = np.polyfit(np.log10(abscissae), np.log10(errors), 1)
    return 10**coefficients[-1], coefficients[-2]


class EOCRecorder:
    """
    .. automethod:: add_data_point

    .. automethod:: estimate_order_of_convergence
    .. automethod:: order_estimate
    .. automethod:: max_error

    .. automethod:: pretty_print
    .. automethod:: write_gnuplot_file
    """

    def __init__(self) -> None:
        self.history: list[tuple[float, float]] = []

    def add_data_point(self, abscissa: float, error: float) -> None:
        if not (isinstance(abscissa, numbers.Number)
                or (isinstance(abscissa, np.ndarray) and abscissa.shape == ())):
            raise TypeError(
                    f"'abscissa' is not a scalar: '{type(abscissa).__name__}'")

        if not (isinstance(error, numbers.Number)
                or (isinstance(error, np.ndarray) and error.shape == ())):
            raise TypeError(f"'error' is not a scalar: '{type(error).__name__}'")

        self.history.append((abscissa, error))

    def estimate_order_of_convergence(self,
            gliding_mean: int | None = None,
            ) -> np.ndarray:
        abscissae = np.array([a for a, e in self.history])
        errors = np.array([e for a, e in self.history])

        # NOTE: in case any of the errors are exactly 0.0, which
        # can give NaNs in `estimate_order_of_convergence`
        emax: float = np.amax(errors)
        errors += (1 if emax == 0 else emax) * np.finfo(errors.dtype).eps

        size = len(abscissae)
        if gliding_mean is None:
            gliding_mean = size

        data_points = size - gliding_mean + 1
        result: np.ndarray = np.zeros((data_points, 2), float)
        for i in range(data_points):
            result[i, 0], result[i, 1] = estimate_order_of_convergence(
                abscissae[i:i+gliding_mean], errors[i:i+gliding_mean])
        return result

    def order_estimate(self) -> float:
        return self.estimate_order_of_convergence()[0, 1]

    def max_error(self) -> float:
        return max(err for absc, err in self.history)

    def _to_table(self, *,
            abscissa_label="h",
            error_label="Error",
            gliding_mean=2,
            abscissa_format="%s",
            error_format="%s",
            eoc_format="%s"):
        from pytools import Table

        tbl = Table()
        tbl.add_row((abscissa_label, error_label, "Running EOC"))

        gm_eoc = self.estimate_order_of_convergence(gliding_mean)
        for i, (absc, err) in enumerate(self.history):
            absc_str = abscissa_format % absc
            err_str = error_format % err
            if i < gliding_mean-1:
                eoc_str = ""
            else:
                eoc_str = eoc_format % (gm_eoc[i - gliding_mean + 1, 1])

            tbl.add_row((absc_str, err_str, eoc_str))

        if len(self.history) > 1:
            order = self.estimate_order_of_convergence()[0, 1]
            tbl.add_row(("Overall", "", eoc_format % order))

        return tbl

    def pretty_print(self, *,
            abscissa_label: str = "h",
            error_label: str = "Error",
            gliding_mean: int = 2,
            abscissa_format: str = "%s",
            error_format: str = "%s",
            eoc_format: str = "%s",
            table_type: str = "markdown") -> str:
        tbl = self._to_table(
                abscissa_label=abscissa_label, error_label=error_label,
                abscissa_format=abscissa_format,
                error_format=error_format,
                eoc_format=eoc_format,
                gliding_mean=gliding_mean)

        if table_type == "markdown":
            return tbl.github_markdown()
        if table_type == "latex":
            return tbl.latex()
        if table_type == "ascii":
            return str(tbl)
        if table_type == "csv":
            return tbl.csv()
        raise ValueError(f"unknown table type: {table_type}")

    @override
    def __str__(self):
        return self.pretty_print()

    def write_gnuplot_file(self, filename: str) -> None:
        with open(filename, "w") as outfile:
            outfile.writelines(f"{absc:f} {err:f}\n" for absc, err in self.history)
            result = self.estimate_order_of_convergence()
            const = result[0, 0]
            order = result[0, 1]
            outfile.write("\n")
            outfile.writelines(
                f"{absc:f} {const * absc**(-order):f}\n"
                for absc, _err in self.history)


def stringify_eocs(*eocs: EOCRecorder,
        names: tuple[str, ...] | None = None,
        abscissa_label: str = "h",
        error_label: str = "Error",
        gliding_mean: int = 2,
        abscissa_format: str = "%s",
        error_format: str = "%s",
        eoc_format: str = "%s",
        table_type: str = "markdown") -> str:
    """
    :arg names: a :class:`tuple` of names to use for the *error_label* of each
        *eoc*.
    """
    if names is not None and len(names) < len(eocs):
        raise ValueError(
                f"insufficient names: got {len(names)} names for "
                f"{len(eocs)} EOCRecorder instances")

    if names is None:
        names = tuple(f"{error_label} {i}" for i in range(len(eocs)))

    from pytools import merge_tables
    tbl = merge_tables(*[eoc._to_table(
        abscissa_label=abscissa_label, error_label=name,
        abscissa_format=abscissa_format,
        error_format=error_format,
        eoc_format=eoc_format,
        gliding_mean=gliding_mean)
        for name, eoc in zip(names, eocs, strict=True)
        ], skip_columns=(0,))

    if table_type == "markdown":
        return tbl.github_markdown()
    if table_type == "latex":
        return tbl.latex()
    if table_type == "ascii":
        return str(tbl)
    if table_type == "csv":
        return tbl.csv()
    raise ValueError(f"unknown table type: {table_type}")

# }}}


# {{{ p convergence verifier

class PConvergenceVerifier:
    def __init__(self):
        self.orders = []
        self.errors = []

    def add_data_point(self, order, error):
        self.orders.append(order)
        self.errors.append(error)

    @override
    def __str__(self):
        from pytools import Table
        tbl = Table()
        tbl.add_row(("p", "error"))

        for p, err in zip(self.orders, self.errors, strict=True):
            tbl.add_row((str(p), str(err)))

        return str(tbl)

    def __call__(self):
        orders = np.array(self.orders, np.float64)
        errors = np.abs(np.array(self.errors, np.float64))

        rel_change = np.diff(1e-20 + np.log10(errors)) / np.diff(orders)

        assert (rel_change < -0.2).all()

# }}}


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = """
Copyright (C) 2009-2019 Andreas Kloeckner
Copyright (C) 2022 University of Illinois Board of Trustees
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """
MPI helper functionality
========================

.. autofunction:: check_for_mpi_relaunch
.. autofunction:: run_with_mpi_ranks
.. autofunction:: pytest_raises_on_rank
"""

import contextlib
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Generator


def check_for_mpi_relaunch(argv):
    if argv[1] != "--mpi-relaunch":
        return

    from pickle import loads
    f, args, kwargs = loads(argv[2])

    f(*args, **kwargs)
    import sys
    sys.exit()


def run_with_mpi_ranks(py_script, ranks, callable_, args=(), kwargs=None):
    if kwargs is None:
        kwargs = {}

    import os
    import sys
    newenv = os.environ.copy()
    newenv["PYTOOLS_RUN_WITHIN_MPI"] = "1"

    from pickle import dumps
    callable_and_args = dumps((callable_, args, kwargs))

    from subprocess import check_call
    check_call(["mpirun", "-np", str(ranks),
        sys.executable, py_script, "--mpi-relaunch", callable_and_args],
        env=newenv)


@contextlib.contextmanager
def pytest_raises_on_rank(
        my_rank: int, fail_rank: int,
        expected_exception: type[BaseException] | tuple[type[BaseException], ...],
        ) -> Generator[contextlib.AbstractContextManager, None, None]:
    """
    Like :func:`pytest.raises`, but only expect an exception on rank *fail_rank*.
    """
    from contextlib import nullcontext

    import pytest

    if my_rank == fail_rank:
        cm: contextlib.AbstractContextManager = pytest.raises(expected_exception)
    else:
        cm = nullcontext()

    with cm as exc:
        yield exc



from __future__ import annotations

import pickle
import sys
from typing import cast

import pytest

import pytools
import pytools.py_codegen as codegen


def test_pickling_with_module_import():
    cg = codegen.PythonCodeGenerator()
    cg("import pytools")
    cg("import math as m")

    import pickle
    mod = pickle.loads(pickle.dumps(cg.get_picklable_module()))

    assert mod.mod_globals["pytools"] is pytools

    import math
    assert mod.mod_globals["m"] is math


def test_picklable_function():
    cg = codegen.PythonFunctionGenerator("f", args=())
    cg("return 1")

    import pickle
    f = pickle.loads(pickle.dumps(cg.get_picklable_function()))

    assert f() == 1


def test_function_decorators(capfd):
    cg = codegen.PythonFunctionGenerator("f", args=(), decorators=["@staticmethod"])
    cg("return 42")

    assert cg.get_function()() == 42

    cg = codegen.PythonFunctionGenerator("f", args=(), decorators=["@classmethod"])
    cg("return 42")

    with pytest.raises(TypeError):
        cg.get_function()()

    cg = codegen.PythonFunctionGenerator("f", args=(),
                                         decorators=["@staticmethod", "@classmethod"])
    cg("return 42")

    with pytest.raises(TypeError):
        cg.get_function()()

    cg = codegen.PythonFunctionGenerator("f", args=("x"),
                decorators=["from functools import lru_cache", "@lru_cache"])
    cg("print('Hello World!')")
    cg("return 42")

    f = cg.get_function()

    assert f(0) == 42
    out, _err = capfd.readouterr()
    assert out == "Hello World!\n"

    assert f(0) == 42
    out, _err = capfd.readouterr()
    assert out == ""  # second print is not executed due to lru_cache


def test_linecache_func() -> None:
    cg = codegen.PythonFunctionGenerator("f", args=())
    cg("return 42")

    func = cg.get_function()
    func()

    mod_name = func.__code__.co_filename

    import linecache

    assert linecache.getlines(mod_name) == [
        "def f():\n",
        "    return 42\n",
    ]

    assert linecache.getline(mod_name, 1) == "def f():\n"
    assert linecache.getline(mod_name, 2) == "    return 42\n"

    pkl = pickle.dumps(cg.get_picklable_function())

    pf = cast("codegen.PicklableFunction", pickle.loads(pkl))

    post_pickle_mod_name = pf._callable.__code__.co_filename

    assert post_pickle_mod_name != mod_name
    assert linecache.getlines(post_pickle_mod_name) == [
        "def f():\n",
        "    return 42\n",
    ]


def test_linecache_mod() -> None:
    cg2 = codegen.PythonCodeGenerator()
    cg2("def f():")
    cg2("    return 37")

    mod = cg2.get_module()
    mod["f"]()
    mod_name = cast("str", mod["__code__"].co_filename)

    assert mod_name

    import linecache
    assert linecache.getlines(mod_name) == [
        "def f():\n",
        "    return 37\n",
    ]


if __name__ == "__main__":
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])



from __future__ import annotations

import time

from pytools import DependentDictionary, Reference


class StopWatch:
    def __init__(self) -> None:
        self.Elapsed: float = 0
        self.LastStart: float | None = None

    def start(self) -> StopWatch:
        assert self.LastStart is None

        self.LastStart = time.time()
        return self

    def stop(self) -> StopWatch:
        assert self.LastStart is not None

        self.Elapsed += time.time() - self.LastStart
        self.LastStart = None
        return self

    def elapsed(self) -> float:
        if self.LastStart:
            return time.time() - self.LastStart + self.Elapsed
        return self.Elapsed


class Job:
    def __init__(self, name: str) -> None:
        self.Name: str = name
        self.StopWatch: StopWatch = StopWatch().start()

        if self.is_visible():
            print(f"{name}...")

    def done(self) -> None:
        elapsed = self.StopWatch.elapsed()

        JOB_TIMES[self.Name] += elapsed
        if self.is_visible():
            print(" " * (len(self.Name) + 2), elapsed, "seconds")

    def is_visible(self) -> bool:
        if PRINT_JOBS.get():
            return self.Name not in HIDDEN_JOBS
        return self.Name in VISIBLE_JOBS


class EtaEstimator:
    def __init__(self, total_steps: int) -> None:
        self.stopwatch: StopWatch = StopWatch().start()
        self.total_steps: int = total_steps
        assert total_steps > 0

    def estimate(self, done: int) -> float | None:
        fraction_done = done / self.total_steps
        time_spent = self.stopwatch.elapsed()

        if fraction_done > 1.0e-5:
            return time_spent / fraction_done - time_spent
        return None


def print_job_summary() -> None:
    for key, value in JOB_TIMES.iteritems():
        print(key, " " * (50 - len(key)), value)


HIDDEN_JOBS: list[str] = []
VISIBLE_JOBS: list[str] = []
JOB_TIMES = DependentDictionary[str, float](lambda d, x: 0.0)
PRINT_JOBS = Reference(True)



from __future__ import annotations


__copyright__ = """
Copyright (C) 2009-2013 Andreas Kloeckner
Copyright (C) 2013- University of Illinois Board of Trustees
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import re
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Literal, NoReturn, TypeAlias, cast

from typing_extensions import Self, final, overload, override


@final
class RuleError(RuntimeError):
    def __init__(self, rule: LexRule) -> None:
        RuntimeError.__init__(self)
        self.Rule = rule

    @override
    def __str__(self) -> str:
        return repr(self.Rule)


@final
class InvalidTokenError(RuntimeError):
    def __init__(self, s: str, str_index: int) -> None:
        RuntimeError.__init__(self)
        self.string = s
        self.index = str_index

    @override
    def __str__(self) -> str:
        return "at index {}: ...{}...".format(
                self.index, self.string[self.index:self.index+20])


@final
class ParseError(RuntimeError):
    def __init__(self, msg: str, s: str, token: Lexed | None) -> None:
        RuntimeError.__init__(self)
        self.message = msg
        self.string = s
        self.Token = token

    @override
    def __str__(self) -> str:
        if self.Token is None:
            return f"{self.message} at end of input"
        return "{} at index {}: ...{}...".format(
                self.message, self.Token[2],
                self.string[self.Token[2]:self.Token[2]+20])


@final
class RE:
    def __init__(self, s: str, flags: int = 0) -> None:
        self.Content = s
        self.RE = re.compile(s, flags)

    @override
    def __repr__(self) -> str:
        return f"RE({self.Content})"


LexRule: TypeAlias = tuple[str, RE | tuple[str | RE, ...]]
LexTable: TypeAlias = Sequence[LexRule]
BasicLexed: TypeAlias = tuple[str, str, int]
LexedWithMatch: TypeAlias = tuple[str, str, int, re.Match[str]]
Lexed: TypeAlias = BasicLexed | LexedWithMatch


def _matches_rule(
            rule: RE | str | tuple[str | RE, ...],
            s: str,
            start: int,
            rule_dict: Mapping[str, RE | tuple[str | RE, ...]],
            debug: bool = False
        ) -> tuple[int, re.Match[str] | None]:
    if debug:
        print("Trying", rule, "on", s[start:])

    if isinstance(rule, tuple):
        if rule[0] == "|":
            for subrule in rule[1:]:
                length, match_obj = _matches_rule(
                        subrule, s, start, rule_dict, debug)
                if not length:
                    continue
                return length, match_obj
        else:
            my_match_length = 0
            for subrule in rule:
                length, _ = _matches_rule(
                        subrule, s, start, rule_dict, debug)
                if not length:
                    break
                my_match_length += length
                start += length
            else:
                return my_match_length, None
        return 0, None

    if isinstance(rule, str):
        return _matches_rule(rule_dict[rule], s, start, rule_dict, debug)

    if isinstance(rule, RE):
        match_obj = rule.RE.match(s, start)
        if match_obj:
            return match_obj.end()-start, match_obj
        return 0, None

    raise RuleError(rule)


@overload
def lex(
            lex_table: LexTable,
            s: str,
            *,
            debug: bool = ...,
            match_objects: Literal[False] = False,
        ) -> Sequence[BasicLexed]:
    ...


@overload
def lex(
            lex_table: LexTable,
            s: str,
            *,
            debug: bool = ...,
            match_objects: Literal[True],
        ) -> Sequence[LexedWithMatch]:
    ...


def lex(
            lex_table: LexTable,
            s: str,
            *,
            debug: bool = False,
            match_objects: bool = False
        ) -> Sequence[Lexed]:
    rule_dict = dict(lex_table)
    result: list[Lexed] = []
    i = 0
    while i < len(s):
        for name, rule in lex_table:
            length, match_obj = _matches_rule(rule, s, i, rule_dict, debug)
            if length:
                if match_objects:
                    assert match_obj
                    result.append((name, s[i:i+length], i, match_obj))
                else:
                    result.append((name, s[i:i+length], i))
                i += length
                break
        else:
            raise InvalidTokenError(s, i)
    return result


@dataclass
class LexIterator:
    lexed: Sequence[tuple[str, str, int] | tuple[str, str, int, re.Match[str]]]
    raw_string: str
    index: int = 0

    def copy(self) -> Self:
        return type(self)(self.lexed, self.raw_string, self.index)

    def assign(self, rhs: LexIterator) -> None:
        assert self.lexed is rhs.lexed
        assert self.raw_string is rhs.raw_string

        self.index = rhs.index

    def next_tag(self, i: int = 0) -> str:
        return self.lexed[self.index + i][0]

    def next_str(self, i: int = 0) -> str:
        return self.lexed[self.index + i][1]

    def next_match_obj(self) -> re.Match[str]:
        _tok, _s, _i, match = cast("LexedWithMatch", self.lexed[self.index])
        return match

    def next_str_and_advance(self) -> str:
        result = self.next_str()
        self.advance()
        return result

    def advance(self) -> None:
        self.index += 1

    def is_at_end(self, i: int = 0) -> bool:
        return self.index + i >= len(self.lexed)

    def is_next(self, tag: str, i: int = 0) -> bool:
        return (
                self.index + i < len(self.lexed)
                and self.next_tag(i) is tag)

    def raise_parse_error(self, msg: str) -> NoReturn:
        if self.is_at_end():
            raise ParseError(msg, self.raw_string, None)

        raise ParseError(msg, self.raw_string, self.lexed[self.index])

    def expected(self, what_expected: str) -> NoReturn:
        if self.is_at_end():
            self.raise_parse_error(
                    f"{what_expected} expected, end of input found instead")
        else:
            self.raise_parse_error(
                    f"{what_expected} expected, {self.next_tag()} found instead")

    def expect_not_end(self) -> None:
        if self.is_at_end():
            self.raise_parse_error("unexpected end of input")

    def expect(self, tag: str) -> None:
        self.expect_not_end()
        if not self.is_next(tag):
            self.expected(str(tag))



from __future__ import annotations


__copyright__ = """
Copyright (C) 2013 Andreas Kloeckner
Copyright (C) 2014 Matt Wala
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """
Dot helper functions
====================

.. autofunction:: dot_escape
.. autofunction:: show_dot
"""

import html
import logging
import os
from pathlib import Path


logger = logging.getLogger(__name__)


# {{{ graphviz / dot interactive show

def dot_escape(s: str) -> str:
    """
    Escape the string *s* for compatibility with the
    `dot <http://graphviz.org/>`__ language, particularly
    backslashes and HTML tags.

    :arg s: The input string to escape.

    :returns: *s* with special characters escaped.
    """
    # "\" and HTML are significant in graphviz.
    return html.escape(s.replace("\\", "\\\\"))


def show_dot(dot_code: str, output_to: str | None = None) -> Path | None:
    """
    Visualize the graph represented by *dot_code*.

    :arg dot_code: An instance of :class:`str` in the `dot <http://graphviz.org/>`__
        language to visualize.
    :arg output_to: An instance of :class:`str` that can be one of:

        - ``"xwindow"`` to visualize the graph as an
          `X window <https://en.wikipedia.org/wiki/X_Window_System>`_.
        - ``"browser"`` to visualize the graph as an SVG file in the
          system's default web-browser.
        - ``"svg"`` to store the dot code as an SVG file on the file system.
          Returns the path to the generated SVG file.

        Defaults to ``"xwindow"`` if X11 support is present, otherwise defaults
        to ``"browser"``.

    :returns: Depends on *output_to*. If ``"svg"``, returns the path to the
        generated SVG file, otherwise returns ``None``.
    """

    import subprocess
    from tempfile import mkdtemp
    temp_path = Path(mkdtemp(prefix="tmp_pytools_dot"))

    dot_file_name = "code.dot"

    (temp_path / dot_file_name).write_text(dot_code)

    # {{{ preprocess 'output_to'

    if output_to is None:
        with subprocess.Popen(["dot", "-T?"],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE
                              ) as proc:
            assert proc.stderr, ("Could not execute the 'dot' program. "
                                 "Please install the 'graphviz' package and "
                                 "make sure it is in your $PATH.")
            supported_formats = proc.stderr.read().decode()

            if " x11 " in supported_formats and "DISPLAY" in os.environ:
                output_to = "xwindow"
            else:
                output_to = "browser"

    # }}}

    if output_to == "xwindow":
        subprocess.check_call(["dot", "-Tx11", dot_file_name], cwd=temp_path)
    elif output_to in ["browser", "svg"]:
        svg_file_name = "code.svg"
        subprocess.check_call(["dot", "-Tsvg", "-o", svg_file_name, dot_file_name],
                              cwd=temp_path)

        full_svg_file_name = temp_path / svg_file_name
        logger.info("show_dot: svg written to '%s'", full_svg_file_name)

        if output_to == "svg":
            return full_svg_file_name
        assert output_to == "browser"

        from webbrowser import open as browser_open
        browser_open(full_svg_file_name.as_uri())
    else:
        raise ValueError("`output_to` can be one of 'xwindow', 'browser', or 'svg',"
                         f" got '{output_to}'")

    return None

# }}}


# vim: foldmethod=marker



from __future__ import annotations

from importlib import metadata
from urllib.request import urlopen


_conf_url = "https://raw.githubusercontent.com/inducer/sphinxconfig/main/sphinxconfig.py"
with urlopen(_conf_url) as _inf:
    exec(compile(_inf.read(), _conf_url, "exec"), globals())

copyright = "2009-21, Andreas Kloeckner"
author = "Andreas Kloeckner"
release = metadata.version("pytools")
version = ".".join(release.split(".")[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "loopy": ("https://documen.tician.de/loopy", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "platformdirs": ("https://platformdirs.readthedocs.io/en/latest", None),
    "pymbolic": ("https://documen.tician.de/pymbolic", None),
    "pytest": ("https://docs.pytest.org/en/stable", None),
    "python": ("https://docs.python.org/3", None),
    "setuptools": ("https://setuptools.pypa.io/en/latest", None),
}

nitpicky = True
autodoc_type_aliases = {
    "GraphT": "pytools.graph.GraphT",
    "NodeT": "pytools.graph.NodeT",
}

nitpick_ignore_regex = [
    ["py:class", r"optype.*"],
]

sphinxconfig_missing_reference_aliases = {
    # numpy typing
    "NDArray": "obj:numpy.typing.NDArray",
    "np.dtype": "class:numpy.dtype",
    "np.ndarray": "class:numpy.ndarray",
    "np.floating": "class:numpy.floating",
    "np.generic[Any]": "class:numpy.generic",
    # pytools typing
    "ObjectArray1D": "obj:pytools.obj_array.ObjectArray1D",
    "ReadableBuffer": "data:pytools.ReadableBuffer",
}


def setup(app):
    app.connect("missing-reference", process_autodoc_missing_reference)  # noqa: F821



# pyright: reportNoOverloadImplementation=none, reportUninitializedInstanceVariable=none

"""
Handling :mod:`numpy` Object Arrays
===================================

.. autoclass:: T_co
.. autoclass:: ResultT
.. autoclass:: ShapeT
.. autoclass:: NumpyTypeT

.. autoclass:: ObjectArray
.. autoclass:: ObjectArray0D
.. autoclass:: ObjectArray1D
.. autoclass:: ObjectArray2D
.. autoclass:: ObjectArrayND

Creation
--------

.. autofunction:: from_numpy
.. autofunction:: to_numpy
.. autofunction:: new_1d
.. autofunction:: flat
.. autofunction:: stack
.. autofunction:: concatenate
.. autofunction:: trace

Mapping
-------

.. autofunction:: vectorize
.. autofunction:: vectorize_n_args

Numpy workarounds
-----------------
These functions work around a `long-standing, annoying numpy issue
<https://github.com/numpy/numpy/issues/1740>`__.

.. autofunction:: obj_array_real
.. autofunction:: obj_array_imag
.. autofunction:: obj_array_real_copy
.. autofunction:: obj_array_imag_copy
"""


from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2020 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from functools import partial, update_wrapper, wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)
from warnings import warn

from typing_extensions import Self, Unpack, deprecated, override


# NOTE: Importing this must not require importing numpy, so do not be tempted
# to move the numpy import out of the type checking block.


if TYPE_CHECKING:
    from collections.abc import Callable, Hashable, Iterator, Sequence

    import numpy as np
    from numpy.typing import NDArray


T_co = TypeVar("T_co", covariant=True)

ResultT = TypeVar("ResultT")
ShapeT = TypeVar("ShapeT", bound=tuple[int, ...])
NumpyTypeT = TypeVar("NumpyTypeT", bound="np.generic[Any]")


class _ObjectArrayMetaclass(type):
    @override
    def __instancecheck__(cls, instance: object) -> bool:
        try:
            import numpy as np
        except ModuleNotFoundError:
            return False
        return isinstance(instance, np.ndarray) and instance.dtype == object


class ObjectArray(Generic[ShapeT, T_co], metaclass=_ObjectArrayMetaclass):
    """This is a fake type used to distinguish :class:`numpy.ndarray`
    instances with a *dtype* of *object* for static type checking.
    Unlike :class:`numpy.ndarray`, this "type" can straightforwardly
    keep track of its entry type.
    All instances of this "type" will actually be :class:`numpy.ndarray` at run time.
    Note that this creates a "universe" of object arrays that's separate
    from the usual numpy-generated one. In order to usefully interact
    in a type-safe manner, object arrays should be "converted" to this
    "type".

    Furthermore, :func:`isinstance` has been customized so as to be accurate
    for this "type".

    For now, there is precise typing support for arrays up to two dimensions.

    .. autoattribute:: shape
    .. autoattribute:: dtype

    The usual set of operations on :class:`numpy.ndarray` instances works
    at run time; a subset of these operations are typed. If what you
    need isn't typed, please submit a pull request.

    This type is intended to be immutable, and hence covariant in *T*.
    """
    shape: ShapeT
    dtype: np.dtype[Any]

    if TYPE_CHECKING:
        def __len__(self) -> int: ...

        @property
        def size(self) -> int: ...

        @property
        def T(self) -> Self: ...  # noqa: N802

        @overload
        def __getitem__(self, x: ShapeT, /) -> T_co: ...

        @overload
        def __getitem__(self: ObjectArray1D[T_co], x: int, /) -> T_co: ...

        @overload
        def __getitem__(
                self: ObjectArray2D[T_co],
                x: int,
            /) -> ObjectArray1D[T_co]: ...

        @overload
        def __getitem__(
                self: ObjectArrayND[T_co],
                x: int,
            /) -> ObjectArrayND[T_co] | T_co: ...

        @overload
        def __getitem__(
            self: ObjectArray1D[T_co],
            x: slice, /) -> ObjectArray1D[T_co]: ...

        @overload
        def __getitem__(
            self: ObjectArray2D[T_co],
            x: slice, /) -> ObjectArray2D[T_co]: ...

        @overload
        def __getitem__(
            self: ObjectArray2D[T_co],
            x: tuple[slice, int], /) -> ObjectArray1D[T_co]: ...

        @overload
        def __getitem__(
            self: ObjectArray2D[T_co],
            x: tuple[int, slice], /) -> ObjectArray1D[T_co]: ...

        @overload
        def __getitem__(
                self: ObjectArrayND[T_co],
                x: tuple[int | slice, ...],
            /) -> ObjectArrayND[T_co] | T_co: ...

        @overload
        def __iter__(self: ObjectArray1D[T_co]) -> Iterator[T_co]: ...
        @overload
        def __iter__(self: ObjectArray2D[T_co]) -> Iterator[ObjectArray1D[T_co]]: ...

        def __pos__(self) -> Self: ...
        def __neg__(self) -> Self: ...
        def __abs__(self) -> Self: ...

        # Technically, arithmetic with T prohibits covariance, because T is a sink
        # here. Precisely typing arithmetic is thorny, and I'm willing to take
        # a gamble on this, I think. -AK, July 2025
        @overload
        def __add__(self, other: Self, /) -> Self: ...
        @overload
        def __add__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __add__(self, other: complex, /) -> Self: ...
        @overload
        def __radd__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __radd__(self, other: complex, /) -> Self: ...

        @overload
        def __sub__(self, other: Self, /) -> Self: ...
        @overload
        def __sub__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __sub__(self, other: complex, /) -> Self: ...
        @overload
        def __rsub__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __rsub__(self, other: complex, /) -> Self: ...

        @overload
        def __mul__(self, other: Self, /) -> Self: ...
        @overload
        def __mul__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __mul__(self, other: complex, /) -> Self: ...
        @overload
        def __rmul__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __rmul__(self, other: complex, /) -> Self: ...

        @overload
        def __truediv__(self, other: Self, /) -> Self: ...
        @overload
        def __truediv__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __truediv__(self, other: complex, /) -> Self: ...
        @overload
        def __rtruediv__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __rtruediv__(self, other: complex, /) -> Self: ...

        @overload
        def __pow__(self, other: Self, /) -> Self: ...
        @overload
        def __pow__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __pow__(self, other: complex, /) -> Self: ...
        @overload
        def __rpow__(self, other: T_co, /) -> Self: ...  # pyright: ignore[reportGeneralTypeIssues]
        @overload
        def __rpow__(self, other: complex, /) -> Self: ...

        @overload
        def __matmul__(
                    self: ObjectArray1D[T_co],
                    other: ObjectArray1D[T_co],
                /) -> T_co: ...
        @overload
        def __matmul__(
                    self: ObjectArray2D[T_co],
                    other: ObjectArray1D[T_co],
                /) -> ObjectArray1D[T_co]: ...
        @overload
        def __matmul__(
                    self: ObjectArray1D[T_co],
                    other: ObjectArray2D[T_co],
                /) -> ObjectArray1D[T_co]: ...
        @overload
        def __matmul__(
                    self: ObjectArray2D[T_co],
                    other: ObjectArray2D[T_co],
                /) -> ObjectArray2D[T_co]: ...
        @overload
        def __matmul__(
                    self: ObjectArray2D[T_co],
                    other: np.ndarray[tuple[int, int], np.dtype[Any]],
                /) -> ObjectArray2D[T_co]: ...
        @overload
        def __matmul__(
                    self: ObjectArrayND[T_co],
                    other: ObjectArrayND[T_co],
                /) -> ObjectArrayND[T_co] | T_co: ...

        @property
        def flat(self) -> Iterator[T_co]: ...

        def flatten(self) -> ObjectArray1D[T_co]: ...

        @overload
        def tolist(self: ObjectArray0D[T_co]) -> T_co: ...
        @overload
        def tolist(self: ObjectArray1D[T_co]) -> list[T_co]: ...
        @overload
        def tolist(self: ObjectArray2D[T_co]) -> list[list[T_co]]: ...
        @overload
        def tolist(self) -> list[Any]: ...


ObjectArray0D: TypeAlias = ObjectArray[tuple[()], T_co]
ObjectArray1D: TypeAlias = ObjectArray[tuple[int], T_co]
ObjectArray2D: TypeAlias = ObjectArray[tuple[int, int], T_co]
ObjectArrayND: TypeAlias = ObjectArray[tuple[int, ...], T_co]


def to_numpy(ary: ObjectArray[ShapeT, T_co]) -> np.ndarray[ShapeT, Any]:
    return cast("np.ndarray[ShapeT, Any]", cast("object", ary))


@overload
def from_numpy(
            ary: np.ndarray[ShapeT, Any],
            tp: None = None,
            /
        ) -> ObjectArray[ShapeT, Any]: ...

@overload
def from_numpy(
            ary: np.ndarray[ShapeT, Any],
            tp: type[T_co],
            /
        ) -> ObjectArray[ShapeT, T_co]: ...


def from_numpy(
            ary: np.ndarray[ShapeT, Any],
            tp: type[T_co] | None = None,  # pyright: ignore[reportUnusedParameter]
            /
        ) -> ObjectArray[ShapeT, T_co]:
    if ary.dtype != object:  # pyright: ignore[reportAny]
        ary = ary.astype(object)
    return cast("ObjectArray[ShapeT, T_co]", cast("object", ary))


@overload
def new_1d(  # pyright: ignore[reportOverlappingOverload]
        res_list: np.ndarray[tuple[int], np.dtype[NumpyTypeT]]
    ) -> ObjectArray1D[NumpyTypeT]: ...

@overload
def new_1d(
        res_list: np.ndarray[tuple[int, int], np.dtype[NumpyTypeT]]
    ) -> ObjectArray1D[np.ndarray[tuple[int], np.dtype[NumpyTypeT]]]: ...

@overload
def new_1d(
        res_list: np.ndarray[tuple[int, int, int], np.dtype[NumpyTypeT]]
    ) -> ObjectArray1D[np.ndarray[tuple[int, int], np.dtype[NumpyTypeT]]]: ...

@overload
def new_1d(
        res_list: np.ndarray[tuple[int, ...], np.dtype[NumpyTypeT]]
    ) -> ObjectArray1D[np.ndarray[tuple[int, ...], np.dtype[NumpyTypeT]]]: ...

@overload
def new_1d(res_list: Sequence[T_co]) -> ObjectArray1D[T_co]: ...


def new_1d(
        res_list: (
            Sequence[T_co]
            | np.ndarray[tuple[int, ...], np.dtype[NumpyTypeT]]
            )
    ) -> (ObjectArray1D[T_co]
          | ObjectArray1D[np.ndarray[tuple[int, ...], np.dtype[NumpyTypeT]]]):
    """Create a one-dimensional object array from *res_list*.
    This differs from ``numpy.array(res_list, dtype=object)``
    by whether it tries to determine its shape by descending
    into nested array-like objects. Consider the following example:

    .. doctest::

        >>> import numpy as np
        >>> a = np.array([np.arange(5), np.arange(5)], dtype=object)
        >>> a
        array([[0, 1, 2, 3, 4],
               [0, 1, 2, 3, 4]], dtype=object)
        >>> a.shape
        (2, 5)
        >>> # meanwhile:
        >>> from pytools.obj_array import new_1d
        >>> b = new_1d([np.arange(5), np.arange(5)])
        >>> b
        array([array([0, 1, 2, 3, 4]), array([0, 1, 2, 3, 4])], dtype=object)
        >>> b.shape
        (2,)

    In some settings (such as when the sub-arrays are large and/or
    live on a GPU), the recursive behavior of :func:`numpy.array`
    can be undesirable.

    .. versionadded:: 2025.2.2

        Renamed from ``make_obj_array``.
    """
    import numpy as np
    result = np.empty((len(res_list),), dtype=object)

    # 'result[:] = res_list' may look tempting, however:
    # https://github.com/numpy/numpy/issues/16564
    for idx in range(len(res_list)):
        result[idx] = res_list[idx]

    return cast("ObjectArray1D[T_co]", cast("object", result))


@deprecated("use obj_array.new_1d instead")
def make_obj_array(res_list: Sequence[T_co]) -> ObjectArray1D[T_co]:
    # No run-time warning yet to avoid excessive warning spam.
    return new_1d(res_list)


def stack(
            arrays: Sequence[ObjectArray[ShapeT, T_co]],
            /, *, axis: Literal[0] = 0,
        ) -> ObjectArray[tuple[int, Unpack[ShapeT]], T_co]:
    """
    .. versionadded:: 2025.2.2
    """
    if axis != 0:
        raise NotImplementedError("axis != 0")

    import numpy as np
    return cast("ObjectArray[tuple[int, Unpack[ShapeT]], T_co]", cast("object",
                            np.stack(cast("Sequence[NDArray[Any]]", arrays))))


def concatenate(
            arrays: Sequence[ObjectArray[tuple[int, Unpack[ShapeT]], T_co]],
            /, *, axis: Literal[0] = 0,
        ) -> ObjectArray[tuple[int, Unpack[ShapeT]], T_co]:
    """
    .. versionadded:: 2025.2.2
    """
    if axis != 0:
        raise NotImplementedError("axis != 0")

    import numpy as np
    return cast("ObjectArray[tuple[int, Unpack[ShapeT]], T_co]", cast("object",
                np.concatenate(cast("Sequence[NDArray[Any]]", arrays))))


def trace(
            array: ObjectArray2D[T_co], /,
        ) -> T_co:
    """
    .. versionadded:: 2025.2.2
    """
    import numpy as np
    return cast("T_co", np.trace(cast("NDArray[Any]", cast("object", array))))


@overload
def sum(
            array: ObjectArrayND[T_co],
            axis: None,
        ) -> T_co: ...


@overload
def sum(
            array: ObjectArray1D[T_co],
            axis: int,
        ) -> T_co: ...


@overload
def sum(
            array: ObjectArray2D[T_co],
            axis: int,
        ) -> ObjectArray1D[T_co]: ...


def sum(
            array: ObjectArrayND[T_co],
            axis: int | None = None,
        ) -> ObjectArrayND[T_co] | T_co:
    import numpy as np
    return cast("ObjectArrayND[T_co] | T_co", np.sum(
            cast("NDArray[Any]", cast("object", array)),
            axis=axis,
        ))


def to_hashable(ary: ObjectArray[ShapeT, T_co] | Hashable, /) -> Hashable:
    if isinstance(ary, ObjectArray):
        ary = cast("ObjectArray[ShapeT, T_co]", ary)
        return tuple(ary.flat)
    return ary


@deprecated("use obj_array.to_hashable")
def obj_array_to_hashable(ary: ObjectArray[ShapeT, T_co] | Hashable, /) -> Hashable:
    return to_hashable(ary)


def flat(*args: ObjectArray[ShapeT, T_co] | list[T_co] | T_co) -> ObjectArray1D[T_co]:
    """Return a one-dimensional flattened object array consisting of
    elements obtained by 'flattening' *args* as follows:

    - The first axis of any non-subclassed object arrays will be flattened
      into the result.
    - Instances of :class:`list` will be flattened into the result.
    - Any other type will appear in the list as-is.
    """
    import numpy as np
    res_list: list[T_co] = []
    for arg in args:
        if isinstance(arg, list):
            res_list.extend(cast("list[T_co]", arg))

        # Only flatten genuine, non-subclassed object arrays.
        elif isinstance(arg, ObjectArray) and type(arg) is np.ndarray:  # pyright: ignore[reportUnnecessaryComparison,reportUnknownArgumentType]
            res_list.extend(arg.flat)

        else:
            res_list.append(cast("T_co", arg))

    return new_1d(res_list)


@deprecated("use obj_array.flat")
def flat_obj_array(
            *args: ObjectArray[ShapeT, T_co] | list[T_co] | T_co
        ) -> ObjectArray1D[T_co]:
    warn("flat_obj_array is deprecated, use obj_array.flat instead. "
         "This will stop working in 2027.", DeprecationWarning, stacklevel=2)
    return flat(*args)


@overload
def vectorize(
            f: Callable[[T_co], ResultT],
            ary: ObjectArray[ShapeT, T_co],
        ) -> ObjectArray[ShapeT, ResultT]: ...

@overload
def vectorize(
            f: Callable[[T_co], ResultT],
            ary: T_co,
        ) -> ResultT: ...


def vectorize(
            f: Callable[[T_co], ResultT],
            ary: T_co | ObjectArray[ShapeT, T_co],
        ) -> ResultT | ObjectArray[ShapeT, ResultT]:
    """Apply the function *f* to all entries of the object array *ary*.
    Return an object array of the same shape consisting of the return
    values.
    If *ary* is not an object array, return ``f(ary)``.

    .. note ::

        This function exists because :class:`numpy.vectorize` suffers from the same
        issue described under :func:`new_1d`.
    """

    import numpy as np
    if isinstance(ary, ObjectArray):
        result = np.empty_like(ary)
        ary = cast("ObjectArray[ShapeT, T_co]", ary)
        for i in np.ndindex(ary.shape):
            result[i] = f(ary[i])
        return cast("ObjectArray[ShapeT, ResultT]", cast("object", result))

    return f(ary)


@deprecated("use obj_array.vectorize")
def obj_array_vectorize(
            f: Callable[[T_co], ResultT],
            ary: T_co | ObjectArray[ShapeT, T_co],
        ) -> ResultT | ObjectArray[ShapeT, ResultT]:
    warn("obj_array_vectorize is deprecated, use obj_array.vectorize instead. "
         "This will stop working in 2027.", DeprecationWarning, stacklevel=2)
    return vectorize(f, ary)


# FIXME: It'd be nice to do make this more precise (T->T, ObjArray->ObjArray),
# but I don't know how.
def vectorized(
            f: Callable[[T_co], T_co]
        ) -> Callable[
                [T_co | ObjectArray[ShapeT, T_co]],
                T_co | ObjectArray[ShapeT, T_co]]:
    wrapper = partial(vectorize, f)
    update_wrapper(wrapper, f)
    return wrapper  # pyright: ignore[reportReturnType]


@deprecated("use obj_array.vectorized instead")
def obj_array_vectorized(
            f: Callable[[T_co], T_co]
        ) -> Callable[
                [T_co | ObjectArray[ShapeT, T_co]],
                T_co | ObjectArray[ShapeT, T_co]]:
    warn("obj_array_vectorized is deprecated, use obj_array.vectorized instead. "
         "This will stop working in 2027.", DeprecationWarning, stacklevel=2)
    return vectorized(f)


# FIXME: I don't know that this function *can* be precisely typed.
# We could probably handle a few specific nesting levels, but is it worth it?
def rec_vectorize(
            f: Callable[[object], object],
            ary: object | ObjectArray[ShapeT, object]
        ) -> object | ObjectArray[ShapeT, object]:
    """Apply the function *f* to all entries of the object array *ary*.
    Return an object array of the same shape consisting of the return
    values.
    If the elements of *ary* are further object arrays, recurse
    until non-object-arrays are found and then apply *f* to those
    entries.
    If *ary* is not an object array, return ``f(ary)``.

    .. note ::

        This function exists because :class:`numpy.vectorize` suffers from the same
        issue described under :func:`new_1d`.
    """
    if isinstance(ary, ObjectArray):
        import numpy as np
        ary = cast("ObjectArray[ShapeT, object]", ary)
        result = np.empty_like(ary)
        for i in np.ndindex(ary.shape):
            result[i] = rec_vectorize(f, ary[i])
        return cast("ObjectArray[Any, ShapeT]", cast("object", result))

    return f(ary)


@deprecated("use obj_array.rec_vectorize instead")
def rec_obj_array_vectorize(
            f: Callable[[object], object],
            ary: object | ObjectArray[ShapeT, object]
        ) -> object | ObjectArray[ShapeT, object]:
    warn("rec_obj_array_vectorized is deprecated, "
         "use obj_array.rec_vectorized instead. "
         "This will stop working in 2027.", DeprecationWarning, stacklevel=2)
    return rec_vectorize(f, ary)


def rec_obj_array_vectorized(
            f: Callable[[object], ResultT]
        ) -> Callable[
                [object | ObjectArray[ShapeT, object]],
                object | ObjectArray[ShapeT, object]]:
    wrapper = partial(rec_vectorize, f)
    update_wrapper(wrapper, f)
    return wrapper


def vectorize_n_args(f, *args):
    """Apply the function *f* elementwise to all entries of any
    object arrays in *args*. All such object arrays are expected
    to have the same shape (but this is not checked).
    Equivalent to an appropriately-looped execution of::

        result[idx] = f(obj_array_arg1[idx], arg2, obj_array_arg3[idx])

    Return an object array of the same shape as the arguments consisting of the
    return values of *f*.

    .. note ::

        This function exists because :class:`numpy.vectorize` suffers from the same
        issue described under :func:`new_1d`.
    """
    import numpy as np
    oarray_arg_indices = []
    for i, arg in enumerate(args):
        if isinstance(arg, np.ndarray) and arg.dtype.char == "O":
            oarray_arg_indices.append(i)

    if not oarray_arg_indices:
        return f(*args)

    leading_oa_index = oarray_arg_indices[0]

    template_ary = args[leading_oa_index]
    result = np.empty_like(template_ary)
    new_args = list(args)
    for i in np.ndindex(template_ary.shape):
        for arg_i in oarray_arg_indices:
            new_args[arg_i] = args[arg_i][i]
        result[i] = f(*new_args)
    return result


@deprecated("use obj_array.vectorize_n_args")
def obj_array_vectorize_n_args(f, *args):
    warn("obj_array_vectorize_n_args is deprecated, "
         "use obj_array.vectorize_n_args instead. "
         "This will stop working in 2027.", DeprecationWarning, stacklevel=2)
    return vectorize_n_args(f, *args)


def vectorized_n_args(f):
    # Unfortunately, this can't use partial(), as the callable returned by it
    # will not be turned into a bound method upon attribute access.
    # This may happen here, because the decorator *could* be used
    # on methods, since it can "look past" the leading `self` argument.
    # Only exactly function objects receive this treatment.
    #
    # Spec link:
    # https://docs.python.org/3/reference/datamodel.html#the-standard-type-hierarchy
    # (under "Instance Methods", quote as of Py3.9.4)
    # > Also notice that this transformation only happens for user-defined functions;
    # > other callable objects (and all non-callable objects) are retrieved
    # > without transformation.

    @wraps(f)
    def wrapper(*args):
        return vectorize_n_args(f, *args)

    return wrapper


@deprecated("use obj_array.vectorized_n_args")
def obj_array_vectorized_n_args(f):
    warn("obj_array_vectorized_n_args is deprecated, "
         "use obj_array.vectorized_n_args instead. "
         "This will stop working in 2027.", DeprecationWarning, stacklevel=2)
    return vectorized_n_args(f)


# {{{ workarounds for https://github.com/numpy/numpy/issues/1740

def obj_array_real(ary):
    return rec_vectorize(lambda x: x.real, ary)


def obj_array_imag(ary):
    return rec_vectorize(lambda x: x.imag, ary)


def obj_array_real_copy(ary):
    return rec_vectorize(lambda x: x.real.copy(), ary)


def obj_array_imag_copy(ary):
    return rec_vectorize(lambda x: x.imag.copy(), ary)

# }}}


# vim: foldmethod=marker



"""OpenMPI, once initialized, prohibits forking. This helper module
allows the forking of *one* helper child process before OpenMPI
initialization that can do the forking for the fork-challenged
parent process.

Since none of this is MPI-specific, it got parked in :mod:`pytools`.

.. autoexception:: ExecError
    :show-inheritance:

.. autoclass:: Forker
.. autoclass:: DirectForker
.. autoclass:: IndirectForker

.. autofunction:: enable_prefork
.. autofunction:: call
.. autofunction:: call_async
.. autofunction:: call_capture_output
.. autofunction:: wait
.. autofunction:: waitall
"""
from __future__ import annotations

import socket
from abc import ABC, abstractmethod
from subprocess import Popen
from typing import TYPE_CHECKING, Any

from typing_extensions import override


if TYPE_CHECKING:
    from collections.abc import Sequence


class ExecError(OSError):
    pass


class Forker(ABC):
    @abstractmethod
    def call(self, cmdline: Sequence[str], cwd: str | None = None) -> int:
        pass

    @abstractmethod
    def call_async(self, cmdline: Sequence[str], cwd: str | None = None) -> int:
        pass

    @abstractmethod
    def call_capture_output(self,
                            cmdline: Sequence[str],
                            cwd: str | None = None,
                            error_on_nonzero: bool = True) -> tuple[int, bytes, bytes]:
        pass

    @abstractmethod
    def wait(self, aid: int) -> int:
        pass

    @abstractmethod
    def waitall(self) -> dict[int, int]:
        pass


class DirectForker(Forker):
    def __init__(self) -> None:
        self.apids: dict[int, Popen[bytes]] = {}
        self.count: int = 0

    @override
    def call(self, cmdline: Sequence[str], cwd: str | None = None) -> int:
        from subprocess import call as spcall

        try:
            return spcall(cmdline, cwd=cwd)
        except OSError as e:
            raise ExecError(
                    "error invoking '{}': {}".format(" ".join(cmdline), e)) from e

    @override
    def call_async(self, cmdline: Sequence[str], cwd: str | None = None) -> int:
        try:
            self.count += 1

            proc = Popen(cmdline, cwd=cwd)
            self.apids[self.count] = proc

            return self.count
        except OSError as e:
            raise ExecError(
                "error invoking '{}': {}".format(" ".join(cmdline), e)) from e

    @override
    def call_capture_output(self,
                            cmdline: Sequence[str],
                            cwd: str | None = None,
                            error_on_nonzero: bool = True) -> tuple[int, bytes, bytes]:
        from subprocess import PIPE, Popen

        try:
            popen = Popen(cmdline, cwd=cwd, stdin=PIPE, stdout=PIPE,
                          stderr=PIPE)
            stdout_data, stderr_data = popen.communicate()

            if error_on_nonzero and popen.returncode:
                raise ExecError("status {} invoking '{}': {}".format(
                    popen.returncode,
                    " ".join(cmdline),
                    stderr_data.decode("utf-8", errors="replace")))

            return popen.returncode, stdout_data, stderr_data
        except OSError as e:
            raise ExecError(
                    "error invoking '{}': {}".format(" ".join(cmdline), e)) from e

    @override
    def wait(self, aid: int) -> int:
        proc = self.apids.pop(aid)
        return proc.wait()

    @override
    def waitall(self) -> dict[int, int]:
        rets = {}

        for aid in self.apids:
            rets[aid] = self.wait(aid)

        return rets


def _send_packet(sock: socket.socket, data: object) -> None:
    from pickle import dumps
    from struct import pack

    packet = dumps(data)

    sock.sendall(pack("I", len(packet)))
    sock.sendall(packet)


def _recv_packet(sock: socket.socket,
                 who: str = "Process",
                 partner: str = "other end") -> tuple[object, ...]:
    from struct import calcsize, unpack
    size_bytes_size = calcsize("I")
    size_bytes = sock.recv(size_bytes_size)

    if len(size_bytes) < size_bytes_size:
        raise SystemExit

    size, = unpack("I", size_bytes)

    packet = b""
    while len(packet) < size:
        packet += sock.recv(size)

    from pickle import loads

    result = loads(packet)
    assert isinstance(result, tuple)

    return result


def _fork_server(sock: socket.socket) -> None:
    # Ignore keyboard interrupts, we'll get notified by the parent.
    import signal
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    # Construct a local DirectForker to do the dirty work
    df = DirectForker()

    funcs = {
        "call": df.call,
        "call_async": df.call_async,
        "call_capture_output": df.call_capture_output,
        "wait": df.wait,
        "waitall": df.waitall
    }

    try:
        while True:
            func_name, args, kwargs = _recv_packet(
                sock, who="Prefork server", partner="parent"
            )
            assert isinstance(func_name, str)

            if func_name == "quit":
                df.waitall()
                _send_packet(sock, ("ok", None))
                break
            try:
                result = funcs[func_name](*args, **kwargs)
            # FIXME: Is catching all exceptions the right course of action?
            except Exception as e:  # pylint:disable=broad-except
                _send_packet(sock, ("exception", e))
            else:
                _send_packet(sock, ("ok", result))
    finally:
        sock.close()

    import os
    os._exit(0)


class IndirectForker(Forker):
    def __init__(self, server_pid: int, sock: socket.socket) -> None:
        self.server_pid = server_pid
        self.socket = sock

        import atexit
        atexit.register(self._quit)

    def _remote_invoke(self, name: str, *args: Any, **kwargs: Any) -> object:
        _send_packet(self.socket, (name, args, kwargs))
        status, result = _recv_packet(
            self.socket, who="Prefork client", partner="prefork server"
        )

        if status == "exception":
            assert isinstance(result, Exception)
            raise result

        assert status == "ok"
        return result

    def _quit(self) -> None:
        self._remote_invoke("quit")

        from os import waitpid
        waitpid(self.server_pid, 0)

    @override
    def call(self, cmdline: Sequence[str], cwd: str | None = None) -> int:
        result = self._remote_invoke("call", cmdline, cwd)

        assert isinstance(result, int)
        return result

    @override
    def call_async(self, cmdline: Sequence[str], cwd: str | None = None) -> int:
        result = self._remote_invoke("call_async", cmdline, cwd)

        assert isinstance(result, int)
        return result

    @override
    def call_capture_output(self,
                            cmdline: Sequence[str],
                            cwd: str | None = None,
                            error_on_nonzero: bool = True,
                            ) -> tuple[int, bytes, bytes]:
        return self._remote_invoke("call_capture_output", cmdline, cwd,
                                   error_on_nonzero)

    @override
    def wait(self, aid: int) -> int:
        result = self._remote_invoke("wait", aid)

        assert isinstance(result, int)
        return result

    @override
    def waitall(self) -> dict[int, int]:
        result = self._remote_invoke("waitall")

        assert isinstance(result, dict)
        return result


forker: Forker = DirectForker()


def enable_prefork() -> None:
    global forker

    if isinstance(forker, IndirectForker):
        return

    s_parent, s_child = socket.socketpair()

    from os import fork
    fork_res = fork()

    # Child
    if fork_res == 0:
        s_parent.close()
        _fork_server(s_child)
    # Parent
    else:
        s_child.close()
        forker = IndirectForker(fork_res, s_parent)


def call(cmdline: Sequence[str], cwd: str | None = None) -> int:
    return forker.call(cmdline, cwd)


def call_async(cmdline: Sequence[str], cwd: str | None = None) -> int:
    return forker.call_async(cmdline, cwd)


def call_capture_output(cmdline: Sequence[str],
                        cwd: str | None = None,
                        error_on_nonzero: bool = True) -> tuple[int, bytes, bytes]:
    return forker.call_capture_output(cmdline, cwd, error_on_nonzero)


def wait(aid: int) -> int:
    return forker.wait(aid)


def waitall() -> dict[int, int]:
    return forker.waitall()



from __future__ import annotations


__copyright__ = """
Copyright (C) 2009-2013 Andreas Kloeckner
Copyright (C) 2020 Matt Wala
Copyright (C) 2020 James Stevens
Copyright (C) 2024 Addison Alvey-Blanco
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """
Graph Algorithms
================

.. note::

    These functions are mostly geared towards directed graphs (digraphs).

.. autofunction:: reverse_graph
.. autofunction:: a_star
.. autofunction:: compute_sccs
.. autoexception:: CycleError
.. autofunction:: compute_topological_order
.. autofunction:: compute_transitive_closure
.. autofunction:: contains_cycle
.. autofunction:: compute_induced_subgraph
.. autofunction:: as_graphviz_dot
.. autofunction:: validate_graph
.. autofunction:: is_connected
.. autofunction:: undirected_graph_from_edges
.. autofunction:: get_reachable_nodes

Type Variables Used
-------------------

.. class:: NodeT

    Type of a graph node, can be any hashable type.

.. class:: GraphT

    A :class:`collections.abc.Mapping` representing a directed
    graph. The mapping contains one key representing each node in the
    graph, and this key maps to a :class:`collections.abc.Collection` of its
    successor nodes. Note that most functions expect that every graph node
    is included as a key in the graph.
"""

from collections.abc import (
    Callable,
    Collection,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    MutableSet,
)
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, TypeAlias, TypeVar


if TYPE_CHECKING:
    import optype

Node: TypeAlias = Hashable
NodeT = TypeVar("NodeT", bound=Node)

GraphT: TypeAlias = Mapping[NodeT, Collection[NodeT]]


# {{{ reverse_graph

def reverse_graph(graph: GraphT[NodeT]) -> GraphT[NodeT]:
    """
    Reverses a graph *graph*.

    :returns: A :class:`dict` representing *graph* with edges reversed.
    """
    result: dict[NodeT, set[NodeT]] = {}

    for node_key, successor_nodes in graph.items():
        # Make sure every node is in the result even if it has no successors
        result.setdefault(node_key, set())

        for successor in successor_nodes:
            result.setdefault(successor, set()).add(node_key)

    return {k: frozenset(v) for k, v in result.items()}

# }}}


# {{{ a_star

@dataclass(frozen=True)
class _AStarNode(Generic[NodeT]):
    state: NodeT
    parent: _AStarNode[NodeT] | None
    path_cost: float | int


def a_star(
        initial_state: NodeT, goal_state: NodeT, neighbor_map: GraphT[NodeT],
        estimate_remaining_cost: Callable[[NodeT], float] | None = None,
        get_step_cost: Callable[[Any, NodeT], float] | None = None,
        ) -> list[NodeT]:
    """
    With the default cost and heuristic, this amounts to Dijkstra's algorithm.
    """

    from heapq import heappop, heappush

    if get_step_cost is None:
        def default_get_step_cost(_x: NodeT, _y: NodeT, /) -> float:
            return 1.0

        get_step_cost = default_get_step_cost

    if estimate_remaining_cost is None:
        def default_estimate_remaining_cost(x: NodeT, /) -> float:
            if x != goal_state:
                return 1

            return 0

        estimate_remaining_cost = default_estimate_remaining_cost

    inf = float("inf")
    init_remcost = estimate_remaining_cost(initial_state)
    assert init_remcost != inf

    queue = [(init_remcost, _AStarNode(initial_state, parent=None, path_cost=0))]
    visited_states: set[NodeT] = set()

    while queue:
        _, top = heappop(queue)
        visited_states.add(top.state)

        if top.state == goal_state:
            result: list[NodeT] = []
            it: _AStarNode[NodeT] | None = top

            while it is not None:
                result.append(it.state)
                it = it.parent

            return result[::-1]

        for state in neighbor_map[top.state]:
            if state in visited_states:
                continue

            remaining_cost = estimate_remaining_cost(state)
            if remaining_cost == inf:
                continue
            step_cost = get_step_cost(top, state)

            estimated_path_cost = top.path_cost+step_cost+remaining_cost
            heappush(queue,
                (estimated_path_cost,
                    _AStarNode(state, top, path_cost=top.path_cost + step_cost)))

    raise RuntimeError("no solution")

# }}}


# {{{ compute SCCs with Tarjan's algorithm

def compute_sccs(graph: GraphT[NodeT]) -> list[list[NodeT]]:
    to_search = set(graph.keys())
    visit_order: dict[NodeT, int] = {}
    scc_root: dict[NodeT, int] = {}
    sccs: list[list[NodeT]] = []

    while to_search:
        top = next(iter(to_search))
        call_stack: list[tuple[NodeT, Iterator[NodeT], NodeT | None]] = (
            [(top, iter(graph[top]), None)])

        visit_stack: list[NodeT] = []
        visiting: set[NodeT] = set()
        scc: list[NodeT] = []

        while call_stack:
            top, children, last_popped_child = call_stack.pop()

            if top not in visiting:
                # Unvisited: mark as visited, initialize SCC root.
                count = len(visit_order)
                visit_stack.append(top)
                visit_order[top] = count
                scc_root[top] = count
                visiting.add(top)
                to_search.discard(top)

            # Returned from a recursion, update SCC.
            if last_popped_child is not None:
                scc_root[top] = min(
                    scc_root[top],
                    scc_root[last_popped_child])

            for child in children:
                if child not in visit_order:
                    # Recurse.
                    call_stack.extend((
                        (top, children, child),
                        (child, iter(graph[child]), None)))
                    break
                if child in visiting:
                    scc_root[top] = min(
                        scc_root[top],
                        visit_order[child])
            else:
                if scc_root[top] == visit_order[top]:
                    scc = []
                    while visit_stack[-1] != top:
                        scc.append(visit_stack.pop())
                    scc.append(visit_stack.pop())
                    for item in scc:
                        visiting.remove(item)
                    sccs.append(scc)

    return sccs

# }}}


# {{{ compute topological order

class CycleError(Exception):
    """
    Raised when a topological ordering cannot be computed due to a cycle.

    .. autoattribute:: node
    """

    node: Node
    """Node in a directed graph that is part of a cycle."""

    def __init__(self, node: Node) -> None:
        self.node = node
        super().__init__(node)


@dataclass(frozen=True)
class _HeapEntry(Generic[NodeT]):
    """
    Helper class to compare associated keys while comparing the elements in
    heap operations.

    Only needs to define :func:`pytools.graph.__lt__` according to
    <https://github.com/python/cpython/blob/8d21aa21f2cbc6d50aab3f420bb23be1d081dac4/Lib/heapq.py#L135-L138>.
    """
    node: NodeT
    key: optype.CanLt[Any]

    def __lt__(self, other: _HeapEntry[NodeT]) -> bool:
        return self.key < other.key


def compute_topological_order(
        graph: GraphT[NodeT],
        # should not use CanLt[object] because many types can't compare against
        # everything under the sun
        key: Callable[[NodeT], optype.CanLt[Any]] | None = None,
        ) -> list[NodeT]:
    """Compute a topological order of nodes in a directed graph.

    :arg key: A custom key function may be supplied to determine the order in
        break-even cases. Expects a function of one argument that is used to
        extract a comparison key from each node of the *graph*.

    :returns: A :class:`list` representing a valid topological ordering of the
        nodes in the directed graph.

    .. note::

        * Requires the keys of the mapping *graph* to be hashable.
        * Implements `Kahn's algorithm <https://w.wiki/YDy>`__.

    .. versionadded:: 2020.2
    """
    def default_keyfunc(_x: NodeT, /) -> int:
        return 0

    # all nodes have the same keys when not provided
    keyfunc = key if key is not None else default_keyfunc

    from heapq import heapify, heappop, heappush

    order: list[NodeT] = []

    # {{{ compute nodes_to_num_predecessors

    nodes_to_num_predecessors = dict.fromkeys(graph, 0)

    for node in graph:
        for child in graph[node]:
            nodes_to_num_predecessors[child] = (
                    nodes_to_num_predecessors.get(child, 0) + 1)

    # }}}

    total_num_nodes = len(nodes_to_num_predecessors)

    # heap: list of instances of HeapEntry(n) where 'n' is a node in
    # 'graph' with no predecessor. Nodes with no predecessors are the
    # schedulable candidates.
    heap = [_HeapEntry(n, keyfunc(n))
            for n, num_preds in nodes_to_num_predecessors.items()
            if num_preds == 0]
    heapify(heap)

    while heap:
        # pick the node with least key
        node_to_be_scheduled = heappop(heap).node
        order.append(node_to_be_scheduled)

        # discard 'node_to_be_scheduled' from the predecessors of its
        # successors since it's been scheduled
        for child in graph.get(node_to_be_scheduled, ()):
            nodes_to_num_predecessors[child] -= 1
            if nodes_to_num_predecessors[child] == 0:
                heappush(heap, _HeapEntry(child, keyfunc(child)))

    if len(order) != total_num_nodes:
        # any node which has a predecessor left is a part of a cycle
        raise CycleError(next(iter(n for n, num_preds in
            nodes_to_num_predecessors.items() if num_preds != 0)))

    return order

# }}}


# {{{ compute transitive closure

def compute_transitive_closure(
        graph: Mapping[NodeT, MutableSet[NodeT]]) -> GraphT[NodeT]:
    """Compute the transitive closure of a directed graph using Warshall's
    algorithm.

    :arg graph: A :class:`collections.abc.Mapping` representing a directed
        graph. The mapping contains one key representing each node in the
        graph, and this key maps to a :class:`collections.abc.MutableSet` of
        nodes that are connected to the node by outgoing edges. This graph may
        contain cycles. This object must be picklable. Every graph node must
        be included as a key in the graph.

    :returns: The transitive closure of the graph, represented using the same
        data type.

    .. versionadded:: 2020.2
    """
    # Warshall's algorithm

    from copy import deepcopy
    closure = deepcopy(graph)

    # (assumes all graph nodes are included in keys)
    for k in graph:
        for n1 in graph:
            for n2 in graph:
                if k in closure[n1] and n2 in closure[k]:
                    closure[n1].add(n2)

    return closure

# }}}


# {{{ check for cycle

def contains_cycle(graph: GraphT[NodeT]) -> bool:
    """Determine whether a graph contains a cycle.

    :returns: A :class:`bool` indicating whether the graph contains a cycle.

    .. versionadded:: 2020.2
    """

    try:
        compute_topological_order(graph)
        return False
    except CycleError:
        return True

# }}}


# {{{ compute induced subgraph

def compute_induced_subgraph(graph: Mapping[NodeT, set[NodeT]],
                             subgraph_nodes: set[NodeT]) -> GraphT[NodeT]:
    """Compute the induced subgraph formed by a subset of the vertices in a
    graph.

    :arg graph: A :class:`collections.abc.Mapping` representing a directed
        graph. The mapping contains one key representing each node in the
        graph, and this key maps to a :class:`collections.abc.Set` of nodes
        that are connected to the node by outgoing edges.

    :arg subgraph_nodes: A :class:`collections.abc.Set` containing a subset of
        the graph nodes in the graph.

    :returns: A :class:`dict` representing the induced subgraph formed by
        the subset of the vertices included in `subgraph_nodes`.

    .. versionadded:: 2020.2
    """

    new_graph: GraphT[NodeT] = {}
    for node, children in graph.items():
        if node in subgraph_nodes:
            new_graph[node] = children & subgraph_nodes

    return new_graph

# }}}


# {{{ as_graphviz_dot

def as_graphviz_dot(graph: GraphT[NodeT],
                    node_labels: Callable[[NodeT], str] | None = None,
                    edge_labels: Callable[[NodeT, NodeT], str] | None = None,
                    ) -> str:
    """
    Create a visualization of the graph *graph* in the
    `dot <http://graphviz.org/>`__ language.

    :arg node_labels: An optional function that returns node labels
        for each node.

    :arg edge_labels: An optional function that returns edge labels
        for each pair of nodes.

    :returns: A string in the `dot <http://graphviz.org/>`__ language.
    """
    from pytools import UniqueNameGenerator
    id_gen = UniqueNameGenerator(forced_prefix="mynode")

    from pytools.graphviz import dot_escape

    if node_labels is None:
        def default_node_labels(x: NodeT, /) -> str:
            return str(x)

        node_labels = default_node_labels

    if edge_labels is None:
        def default_edge_labels(_x: NodeT, _y: NodeT, /) -> str:
            return ""

        edge_labels = default_edge_labels

    node_to_id: dict[NodeT, str] = {}

    for node, targets in graph.items():
        if node not in node_to_id:
            node_to_id[node] = id_gen()
        for t in targets:
            if t not in node_to_id:
                node_to_id[t] = id_gen()

    # Add nodes
    content = "\n".join(
        [f'{node_to_id[node]} [label="{dot_escape(node_labels(node))}"];'
         for node in node_to_id])

    content += "\n"

    # Add edges
    content += "\n".join(
        [f"{node_to_id[node]} -> {node_to_id[t]} "
         f'[label="{dot_escape(edge_labels(node, t))}"];'
         for (node, targets) in graph.items()
         for t in targets])

    return f"digraph mygraph {{\n{content}\n}}\n"

# }}}


# {{{ validate graph

def validate_graph(graph: GraphT[NodeT]) -> None:
    """
    Validates that all successor nodes of each node in *graph* are keys in
    *graph* itself. Raises a :class:`ValueError` if not.
    """
    seen_nodes: set[NodeT] = set()

    for children in graph.values():
        seen_nodes.update(children)

    if not seen_nodes <= graph.keys():
        raise ValueError(
            f"invalid graph, missing keys: {seen_nodes-graph.keys()}")

# }}}


# {{{ is_connected

def is_connected(graph: GraphT[NodeT]) -> bool:
    """
    Returns whether all nodes in *graph* are connected, ignoring
    the edge direction.

    :returns: A :class:`bool` indicating whether the graph is connected.
    """
    if not graph:
        # https://cs.stackexchange.com/questions/52815/is-a-graph-of-zero-nodes-vertices-connected
        return True

    visited: set[NodeT] = set()

    undirected_graph = {node: set(children) for node, children in graph.items()}

    for node, children in graph.items():
        for child in children:
            undirected_graph[child].add(node)

    def dfs(node: NodeT) -> None:
        visited.add(node)
        for child in undirected_graph[node]:
            if child not in visited:
                dfs(child)

    dfs(next(iter(graph.keys())))

    return visited == graph.keys()  # pyright: ignore[reportUnnecessaryComparison]

# }}}


def undirected_graph_from_edges(
            edges: Iterable[tuple[NodeT, NodeT]],
        ) -> GraphT[NodeT]:
    """
    Constructs an undirected graph using *edges*.

    :arg edges: An :class:`Iterable` of pairs of related :class:`NodeT` s.

    :returns: A :class:`GraphT` that is the undirected graph.
    """
    undirected_graph: dict[NodeT, set[NodeT]] = {}

    for lhs, rhs in edges:
        if lhs == rhs:
            raise TypeError("Found loop in edges,"
                            f" LHS, RHS = {lhs}")

        undirected_graph.setdefault(lhs, set()).add(rhs)
        undirected_graph.setdefault(rhs, set()).add(lhs)

    return undirected_graph


def get_reachable_nodes(
        undirected_graph: GraphT[NodeT],
        source_node: NodeT,
        exclude_nodes: Collection[NodeT] | None = None) -> frozenset[NodeT]:
    """
    Returns a :class:`frozenset` of all nodes in *undirected_graph* that are
    reachable from *source_node*.

    If any node from *exclude_nodes* lies on a path between *source_node* and
    some other node :math:`u` in *undirected_graph* and there are no other
    viable paths, then :math:`u` is considered not reachable from *source_node*.

    In the case where *source_node* is in *exclude_nodes*, then no node is
    reachable from *source_node*, so an empty :class:`frozenset` is returned.
    """
    if exclude_nodes is not None and source_node in exclude_nodes:
        return frozenset()

    nodes_visited: set[NodeT] = set()
    reachable_nodes: set[NodeT] = set()
    nodes_to_visit = {source_node}

    if exclude_nodes is None:
        exclude_nodes = set()

    while nodes_to_visit:
        current_node = nodes_to_visit.pop()
        nodes_visited.add(current_node)

        reachable_nodes.add(current_node)

        neighbors = undirected_graph[current_node]
        nodes_to_visit.update({
            node for node in neighbors
            if node not in nodes_visited and node not in exclude_nodes
        })

    return frozenset(reachable_nodes)


# vim: foldmethod=marker



"""
Spatial Binary Tree
===================

.. autoclass:: Point
.. autoclass:: Element
.. autoclass:: BoundingBox
.. autoclass:: SpatialBinaryTree

.. autoclass:: SpatialBinaryTreeBucket
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TextIO, TypeAlias, cast

import numpy as np


if TYPE_CHECKING:
    from collections.abc import Iterator

Point: TypeAlias = np.ndarray[tuple[int], np.dtype[np.floating]]
Element: TypeAlias = Any
BoundingBox: TypeAlias = tuple[Point, Point]
SpatialBinaryTree: TypeAlias = (
    "SpatialBinaryTreeBucket | tuple[SpatialBinaryTree, SpatialBinaryTree]")


def do_boxes_intersect(bl: BoundingBox, tr: BoundingBox) -> bool:
    (bl1, tr1) = bl
    (bl2, tr2) = tr
    (dimension,) = bl1.shape
    return all(max(bl1[i], bl2[i]) <= min(tr1[i], tr2[i]) for i in range(dimension))


def make_buckets(
        bottom_left: Point,
        top_right: Point,
        allbuckets: list[SpatialBinaryTreeBucket],
        max_elements_per_box: int
    ) -> SpatialBinaryTree:
    (dimensions,) = bottom_left.shape
    half = (top_right - bottom_left) / 2.

    def do(dimension: int, pos: Point) -> SpatialBinaryTree:
        if dimension == dimensions:
            origin = bottom_left + pos * half
            bucket = SpatialBinaryTreeBucket(
                    origin,
                    origin + half,
                    max_elements_per_box=max_elements_per_box)

            allbuckets.append(bucket)
            return bucket

        pos[dimension] = 0
        first = do(dimension + 1, pos)

        pos[dimension] = 1
        second = do(dimension + 1, pos)

        return first, second

    return do(0, np.zeros((dimensions,), dtype=bottom_left.dtype))


class SpatialBinaryTreeBucket:
    """This class represents one bucket in a spatial binary tree.
    It automatically decides whether it needs to create more subdivisions
    beneath itself or not.

    .. autoattribute:: elements

    .. autoattribute:: bottom_left
    .. autoattribute:: top_right
    .. autoattribute:: center
    .. autoattribute:: max_elements_per_box

    .. automethod:: insert
    .. automethod:: generate_matches
    .. automethod:: visualize
    .. automethod:: plot
    """

    elements: list[tuple[Element, BoundingBox]]
    """A list of tuples *(element, bbox)* where bbox is again a tuple
    *(lower_left, upper_right)* of :class:`numpy.ndarray` instances
    satisfying ``(lower_right <= upper_right).all()``.
    """

    bottom_left: Point
    top_right: Point
    center: Point
    max_elements_per_box: int

    buckets: SpatialBinaryTree | None
    all_buckets: list[SpatialBinaryTreeBucket] | None

    def __init__(self,
                 bottom_left: Point,
                 top_right: Point,
                 max_elements_per_box: int | None = None) -> None:
        """
        :param bottom_left: a :mod:`numpy.ndarray` of the minimal coordinates
            of the box being partitioned.
        :param top_right: a :mod:`numpy.ndarray` array of the maximal coordinates
            of the box being partitioned.
        """

        self.bottom_left = bottom_left
        self.top_right = top_right
        self.center = (bottom_left + top_right) / 2

        # As long as buckets is None, there are no subdivisions
        self.elements = []
        self.buckets = None
        self.all_buckets = None

        if max_elements_per_box is None:
            dimensions, = self.bottom_left.shape
            max_elements_per_box = cast("int", 8 * 2**dimensions)

        self.max_elements_per_box = max_elements_per_box

    def insert(self, element: Element, bbox: BoundingBox) -> None:
        """Insert an element into the spatial tree.

        :param element: the element to be stored in the retrieval data
            structure. It is treated as opaque and no assumptions are made on it.

        :param bbox: a bounding box supplied as a tuple ``(bottom_left, top_right)``
            of :mod:`numpy` vectors, such that ``(bottom_left <= top_right).all()``.
            Despite these names, the bounding box (and this entire data structure)
            may be of any dimension.
        """

        def insert_into_subdivision(element: Element, bbox: BoundingBox) -> None:
            assert self.all_buckets is not None

            bucket_matches = [
                ibucket
                for ibucket, bucket in enumerate(self.all_buckets)
                if do_boxes_intersect((bucket.bottom_left, bucket.top_right), bbox)]

            from random import uniform
            if len(bucket_matches) > len(self.all_buckets) // 2:
                # Would go into more than half of all buckets--keep it here
                self.elements.append((element, bbox))
            elif len(bucket_matches) > 1 and uniform(0, 1) > 0.95:
                # Would go into more than one bucket and therefore may recurse
                # indefinitely. Keep it here with a low probability.
                self.elements.append((element, bbox))
            else:
                for ibucket_match in bucket_matches:
                    self.all_buckets[ibucket_match].insert(element, bbox)

        if self.buckets is None:
            # No subdivisions yet.
            if len(self.elements) > self.max_elements_per_box:
                # Too many elements. Need to subdivide.
                self.all_buckets = []
                self.buckets = make_buckets(
                        self.bottom_left, self.top_right,
                        self.all_buckets,
                        max_elements_per_box=self.max_elements_per_box)

                old_elements = self.elements
                self.elements = []

                # Move all elements from the full bucket into the new finer ones
                for el, el_bbox in old_elements:
                    insert_into_subdivision(el, el_bbox)

                insert_into_subdivision(element, bbox)
            else:
                # Simple:
                self.elements.append((element, bbox))
        else:
            # Go find which sudivision to place element
            insert_into_subdivision(element, bbox)

    def generate_matches(self, point: Point) -> Iterator[Element]:
        if self.buckets:
            # We have subdivisions. Use them.
            (dimensions,) = point.shape
            bucket = self.buckets
            for dim in range(dimensions):
                assert isinstance(bucket, tuple)
                bucket = bucket[0] if point[dim] < self.center[dim] else bucket[1]

            assert isinstance(bucket, SpatialBinaryTreeBucket)
            yield from bucket.generate_matches(point)

        # Perform linear search.
        for el, _ in self.elements:
            yield el

    def visualize(self, file: TextIO) -> None:
        file.write(f"{self.bottom_left[0]:f} {self.bottom_left[1]:f}\n")
        file.write(f"{self.top_right[0]:f} {self.bottom_left[1]:f}\n")
        file.write(f"{self.top_right[0]:f} {self.top_right[1]:f}\n")
        file.write(f"{self.bottom_left[0]:f} {self.top_right[1]:f}\n")
        file.write(f"{self.bottom_left[0]:f} {self.bottom_left[1]:f}\n\n")

        if self.buckets:
            assert self.all_buckets is not None
            for i in self.all_buckets:
                i.visualize(file)

    def plot(self, **kwargs: Any) -> None:
        """
        :param ax: an :class:`~matplotlib.axes.Axes` object on which to plot the tree.
        :param kwargs: any remaining arguments are passed to
            :class:`matplotlib.patches.PathPatch`.
        """
        import matplotlib.pyplot as plt
        from matplotlib.patches import PathPatch
        from matplotlib.path import Path

        ax = kwargs.pop("ax", None)
        if ax is None:
            ax = plt.gca()

        el = self.bottom_left
        eh = self.top_right
        pathdata = [
            (Path.MOVETO, (el[0], el[1])),
            (Path.LINETO, (eh[0], el[1])),
            (Path.LINETO, (eh[0], eh[1])),
            (Path.LINETO, (el[0], eh[1])),
            (Path.CLOSEPOLY, (el[0], el[1])),
            ]

        codes, verts = zip(*pathdata, strict=True)
        path = Path(verts, codes)
        patch = PathPatch(path, **kwargs)
        ax.add_patch(patch)

        if self.buckets:
            assert self.all_buckets is not None
            for bucket in self.all_buckets:
                bucket.plot(ax=ax, **kwargs)
